import re
from pathlib import Path
from typing import Callable, Optional, Union

from runhouse.logger import get_logger

from runhouse.resources.functions.function import Function

logger = get_logger(__name__)


def function(
    fn: Optional[Union[str, Callable]] = None,
    name: Optional[str] = None,
    load_from_den: bool = True,
    dryrun: bool = False,
    serialize_notebook_fn: bool = False,
):
    """runhouse.function(fn: str | Callable | None = None, name: str | None = None, system: str | Cluster | None = None, dryrun: bool = False, load_secrets: bool = False, serialize_notebook_fn: bool = False)

    Builds an instance of :class:`Function`.

    Args:
        fn (Optional[str or Callable]): The function to execute on the remote system when the function is called.
        name (Optional[str]): Name of the Function to create or retrieve.
            This can be either from a local config or from Den. (Default: ``None``)
        load_from_den (bool, optional): Whether to try loading the function from Den. (Default: ``True``)
        dryrun (bool, optional): Whether to create the Function if it doesn't exist, or load the Function object as
            a dryrun. (Default: ``False``)
        serialize_notebook_fn (bool, optional): If function is of a notebook setting, whether or not to serialized the
            function. (Default: ``False``)

    Returns:
        Function: The resulting Function object.

    Example:
        >>> cluster = rh.ondemand_cluster(name="my_cluster")
        >>> def sum(a, b):
        >>>    return a + b
        >>>
        >>> summer = rh.function(fn=sum, name="my_func").to(cluster, env=['requirements.txt']).save()
        >>>
        >>> # using the function
        >>> res = summer(5, 8)  # returns 13

        >>> # Load function from above
        >>> reloaded_function = rh.function(name="my_func")
    """  # noqa: E501
    if name and not fn:
        # Try reloading existing function
        return Function.from_name(name, load_from_den=load_from_den, dryrun=dryrun)

    fn_pointers = None
    if callable(fn):
        fn_pointers = Function._extract_pointers(fn)

        if fn_pointers[1] == "notebook":
            fn_pointers = Function._handle_nb_fn(
                fn,
                fn_pointers=fn_pointers,
                serialize_notebook_fn=serialize_notebook_fn,
                name=fn_pointers[2] or name,
            )
    elif isinstance(fn, str):
        # Url must match a regex of the form
        # 'https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name'
        # Use a regex to extract username, repo_name, branch_name, path/to/file.py, and func_name
        pattern = (
            r"https://github\.com/(?P<username>[^/]+)/(?P<repo_name>[^/]+)/blob/"
            r"(?P<branch_name>[^/]+)/(?P<path>[^:]+):(?P<func_name>.+)"
        )
        match = re.match(pattern, fn)

        if match:
            repo_name = match.group("repo_name")
            path = match.group("path")
            func_name = match.group("func_name")
        else:
            raise ValueError(
                "fn must be a callable or string of the form "
                '"https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name"'
            )
        module_name = Path(path).stem
        relative_path = str(repo_name / Path(path).parent)
        fn_pointers = (relative_path, module_name, func_name)
        # TODO [DG] check if the user already added this in their reqs
        # For now, user needs to manually install the corresponding GH repo

    new_function = Function(fn_pointers=fn_pointers, name=name, dryrun=dryrun)

    return new_function



import tempfile
import textwrap
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import requests

from runhouse.globals import rns_client
from runhouse.servers.caddy.config import CaddyConfig
from runhouse.servers.http.http_utils import PutObjectParams, serialize_data


@pytest.mark.servertest
class TestCaddyConfiguration:
    @pytest.fixture(autouse=True)
    def init_fixtures(self):
        # mock existence of cert and key files
        with (patch.object(Path, "exists", return_value=True)):
            self.http_config = CaddyConfig(
                address="127.0.0.1",
            )

            # Using HTTPS client with Caddy handling certs generation
            self.https_config_caddy_certs = CaddyConfig(
                address="127.0.0.1", use_https=True, domain="run.house"
            )

            # Using temporary certs for testing
            with tempfile.NamedTemporaryFile(
                suffix=".cert", delete=False
            ) as cert_file, tempfile.NamedTemporaryFile(
                suffix=".key", delete=False
            ) as key_file:
                self.temp_cert_path = cert_file.name
                self.temp_key_path = key_file.name

            # Using HTTPS client with pre-existing custom certs
            self.https_config_custom_certs = CaddyConfig(
                address="127.0.0.1",
                use_https=True,
                ssl_key_path=self.temp_key_path,
                ssl_cert_path=self.temp_cert_path,
            )

            self.caddy_certs_template = self.https_config_caddy_certs._https_template()
            self.custom_certs_template = (
                self.https_config_custom_certs._https_template()
            )

    # -----------------------------------------

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_caddy_http_build_config(self, mock_subprocess_run):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        self.http_config._build_template()

        template = self.http_config._http_template()
        assert f"http://{self.http_config.address}" in template

        # Assert that subprocess.run was called to change permissions
        mock_subprocess_run.assert_called()

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_https_custom_certs_build_config(self, mock_subprocess_run):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        self.https_config_custom_certs._build_template()

        template = self.https_config_custom_certs._https_template()
        assert template == self.custom_certs_template

        # Assert that subprocess.run was called to change permissions
        mock_subprocess_run.assert_called()

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_https_caddy_certs_build_config(self, mock_subprocess_run):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        self.https_config_caddy_certs._build_template()

        template = self.https_config_caddy_certs._https_template()
        assert template == self.caddy_certs_template

        # Assert that subprocess.run was called to change permissions
        mock_subprocess_run.assert_called()

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_caddy_http_reload(self, mock_subprocess_run):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        expected_command = ["sudo", "systemctl", "reload", "caddy"]

        self.http_config.reload()

        mock_subprocess_run.assert_called_once_with(
            expected_command,
            capture_output=True,
            text=True,
        )

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_caddy_https_reload(self, mock_subprocess_run):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        expected_command = ["sudo", "systemctl", "reload", "caddy"]

        self.https_config_custom_certs.reload()

        mock_subprocess_run.assert_called_once_with(
            expected_command,
            capture_output=True,
            text=True,
        )

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_caddy_reload_error_handling(self, mock_subprocess_run):
        # Simulate a failed reload
        mock_subprocess_run.return_value = MagicMock(
            returncode=1, stderr="Failed to reload"
        )

        with pytest.raises(RuntimeError):
            self.http_config.reload()

    @pytest.mark.level("unit")
    @patch("pathlib.Path.exists")
    @patch("subprocess.run")
    def test_https_firewall_rule_application(
        self, mock_subprocess_run, mock_path_exists
    ):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        mock_path_exists.return_value = True

        self.https_config_caddy_certs._build_template()

        mock_subprocess_run.assert_any_call(
            "sudo ufw allow 443/tcp",
            shell=True,
            check=True,
            capture_output=True,
            text=True,
        )

    @pytest.mark.level("unit")
    def test_invalid_ssl_paths_for_https(self):
        with pytest.raises(FileNotFoundError):
            # Incorrect SSL key and cert paths provided
            CaddyConfig(
                address="127.0.0.1",
                ssl_cert_path="/path/to/nonexistent/cert",
                ssl_key_path="/path/to/nonexistent/key",
                use_https=True,
            )

    @pytest.mark.level("unit")
    def test_ignore_invalid_ssl_paths_for_http(self):
        # Invalid SSL key and cert paths not provided, which we ignore for HTTP config
        CaddyConfig(
            address="127.0.0.1",
            ssl_cert_path="/path/to/nonexistent/cert",
            ssl_key_path="/path/to/nonexistent/key",
        )

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_template_generation_based_on_http_config(self, mock_subprocess_run):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        self.http_config._build_template()

        expected_template = self.http_config._http_template()
        assert expected_template == self.http_config._http_template()

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_template_generation_based_on_https_custom_certs_config(
        self, mock_subprocess_run
    ):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        self.https_config_custom_certs._build_template()

        expected_template = self.https_config_custom_certs._https_template()
        assert expected_template == self.custom_certs_template

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_template_generation_based_on_https_caddy_certs_config(
        self, mock_subprocess_run
    ):
        mock_subprocess_run.return_value = MagicMock(returncode=0)

        self.https_config_caddy_certs._build_template()

        expected_template = self.https_config_caddy_certs._https_template()
        assert expected_template == self.caddy_certs_template

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_caddy_http_config_generation(self, mock_subprocess_run):
        # Set up CaddyConfig for HTTP only
        config = CaddyConfig(address="127.0.0.1")

        mock_subprocess_run.return_value = MagicMock(returncode=0)

        config._build_template()

        # Assert that we wrote the HTTP template to the file
        http_template = config._http_template()

        assert not config.use_https
        assert config.ssl_cert_path is None
        assert config.ssl_key_path is None

        expected_http_template = textwrap.dedent(
            f"""
            {{
                default_sni {config.address}
            }}

            http://{config.address} {{
                reverse_proxy localhost:{config.rh_server_port}
            }}
        """
        ).strip()

        assert http_template == expected_http_template

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_caddy_https_custom_certs_config_generation(self, mock_subprocess_run):
        # Set up CaddyConfig for HTTPS only
        config = CaddyConfig(
            address="127.0.0.1",
            ssl_cert_path=self.temp_cert_path,
            ssl_key_path=self.temp_key_path,
            use_https=True,
        )

        mock_subprocess_run.return_value = MagicMock(returncode=0)

        config._build_template()

        https_template = config._https_template()

        assert config.use_https
        assert config.ssl_cert_path is not None
        assert config.ssl_key_path is not None

        assert https_template == self.custom_certs_template

    @pytest.mark.level("unit")
    @patch("subprocess.run")
    def test_caddy_https_caddy_certs_config_generation(self, mock_subprocess_run):
        # Set up CaddyConfig for HTTPS only
        config = CaddyConfig(address="127.0.0.1", use_https=True, domain="run.house")

        mock_subprocess_run.return_value = MagicMock(returncode=0)

        config._build_template()

        https_template = config._https_template()

        assert config.use_https

        # Caddy will handle cert creation automatically
        assert config.ssl_cert_path is None
        assert config.ssl_key_path is None

        assert https_template == self.caddy_certs_template

    @pytest.mark.level("unit")
    def test_invalid_https_configuration(self):
        # Set up CaddyConfig for HTTPS without specifying valid cert files or a domain
        with pytest.raises(ValueError):
            CaddyConfig(address="127.0.0.1", use_https=True)

        with pytest.raises(FileNotFoundError):
            CaddyConfig(
                address="127.0.0.1",
                use_https=True,
                ssl_cert_path="/some/random/path",
                ssl_key_path="/some/random/other/path",
            )

    @pytest.mark.level("unit")
    def test_use_certs_even_if_domain_provided(self):
        cc = CaddyConfig(
            address="127.0.0.1",
            use_https=True,
            ssl_cert_path="/some/random/path",
            ssl_key_path="/some/random/other/path",
            domain="run.house",
        )
        assert (
            str(cc.ssl_cert_path) == "/some/random/path"
            and str(cc.ssl_key_path) == "/some/random/other/path"
        )
        assert f"tls {cc.ssl_cert_path} {cc.ssl_key_path}" in cc._https_template()

    @pytest.mark.level("unit")
    def test_use_domain_with_no_certs(self):
        cc = CaddyConfig(address="127.0.0.1", use_https=True, domain="run.house")
        assert cc.ssl_cert_path is None and cc.ssl_key_path is None
        assert "https://run.house" in cc._https_template()


@pytest.mark.servertest
class TestCaddyServerLocally:
    UNIT = {
        "cluster": [
            "docker_cluster_pk_http_exposed",
            "docker_cluster_pk_tls_den_auth",
        ]
    }
    LOCAL = {
        "cluster": [
            "docker_cluster_pk_http_exposed",
            "docker_cluster_pk_tls_den_auth",
        ]
    }
    MINIMAL = {
        "cluster": [
            "docker_cluster_pk_http_exposed",
            "docker_cluster_pk_tls_den_auth",
        ]
    }
    RELEASE = {
        "cluster": [
            "docker_cluster_pk_http_exposed",
            "docker_cluster_pk_tls_den_auth",
        ]
    }
    MAXIMAL = {
        "cluster": [
            "docker_cluster_pk_http_exposed",
            "docker_cluster_pk_tls_den_auth",
        ]
    }

    @pytest.mark.level("local")
    def test_using_caddy_on_local_cluster(self, cluster):
        protocol = "https" if cluster._use_https else "http"

        cluster.client.check_server()

        assert cluster.is_up()

        key = "key1"
        test_list = list(range(5, 50, 2)) + ["a string"]
        verify = cluster.client.verify
        response = requests.post(
            f"{protocol}://{cluster.server_address}:{cluster.client_port}/object",
            json=PutObjectParams(
                serialized_data=serialize_data(test_list, "pickle"),
                key=key,
                serialization="pickle",
            ).model_dump(),
            headers=rns_client.request_headers(cluster.rns_address),
            verify=verify,
        )
        assert response.status_code == 200

        response = requests.get(
            f"{protocol}://{cluster.server_address}:{cluster.client_port}/keys",
            headers=rns_client.request_headers(cluster.rns_address),
            verify=verify,
        )

        assert response.status_code == 200
        assert key in response.json().get("data")



import datetime

import ipaddress
import os
import ssl
import threading

from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path

import pytest
import requests
from cryptography import x509
from cryptography.hazmat.backends import default_backend

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.x509 import load_pem_x509_certificate
from cryptography.x509.oid import NameOID
from requests.exceptions import SSLError

from runhouse.rns.utils.api import resolve_absolute_path


class SimpleHTTPRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b"Hello, SSL!")


def create_test_https_server(
    cert_file, key_file, server_class=HTTPServer, handler_class=SimpleHTTPRequestHandler
):
    httpd = server_class(("localhost", 0), handler_class)
    httpd.socket = ssl.wrap_socket(
        httpd.socket, certfile=cert_file, keyfile=key_file, server_side=True
    )
    sa = httpd.socket.getsockname()
    thread = threading.Thread(target=httpd.serve_forever)
    thread.daemon = True
    thread.start()
    return httpd, sa[1], thread


@pytest.mark.servertest
class TestTLSCertConfig:
    @pytest.mark.level("unit")
    def test_generate_certs(self, cert_config):
        assert Path(cert_config.cert_path).exists()
        assert Path(cert_config.key_path).exists()

        # Load the certificate and check properties
        with open(cert_config.cert_path, "rb") as cert_file:
            cert_data = cert_file.read()
            certificate = load_pem_x509_certificate(cert_data, default_backend())
            assert (
                certificate.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
                == "run.house"
            )

            assert ipaddress.IPv4Address(
                "127.0.0.1"
            ) in certificate.extensions.get_extension_for_class(
                x509.SubjectAlternativeName
            ).value.get_values_for_type(
                x509.IPAddress
            )

        # Load the private key and check type
        with open(cert_config.key_path, "rb") as key_file:
            key_data = key_file.read()
            private_key = load_pem_private_key(
                key_data, password=None, backend=default_backend()
            )
            assert isinstance(
                private_key, rsa.RSAPrivateKey
            ), "Private key is not an RSA key."

    @pytest.mark.level("unit")
    def test_resolve_absolute_path(self, mocker, cert_config):

        # set up mocks
        mock_expanduser = mocker.patch("os.path.expanduser")
        mock_abspath = mocker.patch("os.path.abspath")

        # Mock the expanduser and abspath to return a mock path
        mock_expanduser.return_value = "/mocked/home/user/ssl/certs/rh_server.crt"
        mock_abspath.return_value = "/mocked/absolute/path/to/ssl/certs/rh_server.crt"

        resolved_path = resolve_absolute_path(cert_config.cert_path)

        mock_expanduser.assert_any_call(cert_config.cert_path)
        mock_abspath.assert_any_call(mock_expanduser.return_value)

        assert resolved_path == mock_abspath.return_value


@pytest.mark.servertest
class TestHTTPSCertValidity:
    @pytest.fixture(autouse=True)
    def init_fixtures(self):
        self.cert_file = "test_cert.pem"
        self.key_file = "test_key.pem"
        self._generate_test_certificate(self.cert_file, self.key_file)

        # Start a local HTTPS server using the self-signed certificate
        self.server, self.port, self.server_thread = create_test_https_server(
            self.cert_file, self.key_file
        )

        yield

        # Clean up the server and the generated certificate files
        self.server.shutdown()
        self.server.server_close()
        self.server_thread.join()
        os.remove(self.cert_file)
        os.remove(self.key_file)

    def _generate_test_certificate(self, cert_file, key_file):
        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )

        with open(key_file, "wb") as f:
            f.write(
                private_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            )

        subject = issuer = x509.Name(
            [
                x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
                x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "California"),
                x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
                x509.NameAttribute(NameOID.ORGANIZATION_NAME, "My Company"),
                x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
            ]
        )
        cert = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(private_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10))
            .add_extension(
                x509.SubjectAlternativeName([x509.DNSName("localhost")]),
                critical=False,
            )
            .sign(private_key, hashes.SHA256(), default_backend())
        )

        with open(cert_file, "wb") as f:
            f.write(cert.public_bytes(serialization.Encoding.PEM))

    @pytest.mark.level("unit")
    def test_https_request_with_cert_verification(self):
        response = requests.get(f"https://localhost:{self.port}", verify=self.cert_file)

        assert response.status_code == 200
        assert response.text == "Hello, SSL!"

    @pytest.mark.level("unit")
    def test_https_request_with_invalid_cert_verification(self):
        dummy_cert_path = "dummy_cert.pem"
        with open(dummy_cert_path, "w") as f:
            f.write("-----BEGIN CERTIFICATE-----\n")
            f.write("Invalid certificate content\n")
            f.write("-----END CERTIFICATE-----\n")

        # should raise an SSLError because the cert won't be valid for the server's SSL setup
        with pytest.raises(SSLError):
            requests.get(f"https://localhost:{self.port}", verify=dummy_cert_path)

        os.remove(dummy_cert_path)

    @pytest.mark.level("unit")
    def test_https_request_with_self_signed_cert(self):
        response = requests.get(
            f"https://localhost:{self.port}",
            verify=self.cert_file,
        )

        assert response.status_code == 200
        assert response.text == "Hello, SSL!"

    @pytest.mark.level("unit")
    def test_https_request_with_verified_cert(self, mocker):
        # Mock the response of the requests.get call
        mock_get = mocker.patch("requests.get")
        mock_response = requests.Response()
        mock_response.status_code = 200
        mock_response._content = b"Hello, SSL!"
        mock_get.return_value = mock_response

        # Simulate the scenario where the certificate should be verified
        should_verify = True

        # Call the method under test
        response = requests.get(f"https://localhost:{self.port}", verify=should_verify)

        mock_get.assert_called_once_with(
            f"https://localhost:{self.port}", verify=should_verify
        )

        assert response.status_code == 200
        assert response.text == "Hello, SSL!"



import asyncio
import copy
import inspect
import logging
import os
import sys
import time
import uuid
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel

from runhouse.constants import (
    LOGGING_WAIT_TIME,
    LOGS_TO_SHOW_UP_CHECK_TIME,
    MAX_LOGS_TO_SHOW_UP_WAIT_TIME,
    RH_LOGFILE_PATH,
)
from runhouse.logger import get_logger

from runhouse.rns.defaults import req_ctx
from runhouse.rns.utils.api import ResourceVisibility

from runhouse.utils import (
    arun_in_thread,
    generate_default_name,
    is_python_package_string,
    LogToFolder,
    run_with_logs,
    set_env_vars_in_current_process,
    sync_function,
)

logger = get_logger(__name__)


class RaySetupOption(str, Enum):
    GET_OR_FAIL = "get_or_fail"
    TEST_PROCESS = "test_process"


class ClusterServletSetupOption(str, Enum):
    GET_OR_CREATE = "get_or_create"
    GET_OR_FAIL = "get_or_fail"
    FORCE_CREATE = "force_create"


class ObjStoreError(Exception):
    pass


class RunhouseStopIteration(Exception):
    pass


class ActiveFunctionCallInfo(BaseModel):
    key: str
    method_name: str
    request_id: str
    start_time: float
    run_name: str


class NoLocalObjStoreError(ObjStoreError):
    def __init__(self, *args):
        super().__init__("No local object store exists; cannot perform operation.")


def get_cluster_servlet(
    create_if_not_exists: bool = False,
    runtime_env: Optional[Dict] = None,
):
    import ray

    from runhouse.servers.cluster_servlet import ClusterServlet

    if not ray.is_initialized():
        raise ConnectionError("Ray is not initialized.")

    # Previously used list_actors here to avoid a try/except, but it is finicky
    # when there are several Ray clusters running. In tests, we typically run multiple
    # clusters, so let's avoid this.
    try:
        cluster_servlet = ray.get_actor("cluster_servlet", namespace="runhouse")
    except ValueError:
        cluster_servlet = None

    # Ensure the ClusterServlet starts on the head node, per
    # https://discuss.ray.io/t/how-to-ensure-actor-is-running-on-the-same-node-only/2083/3
    current_ip = ray.get_runtime_context().worker.node_ip_address
    if cluster_servlet is None and create_if_not_exists:
        cluster_servlet = (
            ray.remote(ClusterServlet)
            .options(
                name="cluster_servlet",
                get_if_exists=True,
                lifetime="detached",
                namespace="runhouse",
                max_concurrency=1000,
                resources={f"node:{current_ip}": 0.001},
                num_cpus=0,
                runtime_env=runtime_env,
            )
            .remote()
        )

        # Make sure cluster servlet is actually initialized
        ray.get(cluster_servlet.aget_cluster_config.remote())

    return cluster_servlet


def context_wrapper(func):
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        ctx_token = None
        try:
            if not req_ctx.get():
                ctx_token = await self.apopulate_ctx_locally()

            res = await func(self, *args, **kwargs)
        except Exception as e:
            raise e
        finally:
            if ctx_token:
                self.unset_ctx(ctx_token)

        return res

    return wrapper


class ObjStore:
    """Class to handle internal IPC and storage for Runhouse.

    We interact with individual Servlets as well as the global ClusterServlet
    via this class.

    The point of this is that this information can
    be accessed by any node in the cluster, as well as any
    process on any node, via this class.

    1. We store the state of the cluster in the ClusterServlet.
    2. We store an auth cache in the ClusterServlet
    3. We interact with a distributed KV store, which is the most in-depth of these use cases.

        The KV store is used to store objects that are shared across the cluster. Each Servlet
        will have its own ObjStore initialized with a servlet name. This means it will have a
        local Python dictionary with its values. However, each ObjStore can also access the other env
        servlets' KV stores, so we can get and put values across the cluster.

        We maintain individual KV stores in each Servlet's memory so that we can access them in-memory
        if functions within that Servlet make key/value requests.
    """

    def __init__(self):
        self.servlet_name: Optional[str] = None
        self.cluster_servlet: Optional[ray.actor.ActorHandle] = None
        self.cluster_config: Optional[Dict[str, Any]] = None
        self.imported_modules = {}
        self._kv_store: Dict[Any, Any] = None
        self.servlet_cache = {}
        self.active_function_calls = {}
        self.active_bash_run_names = set()

    async def ainitialize(
        self,
        servlet_name: Optional[str] = None,
        has_local_storage: bool = False,
        setup_ray: RaySetupOption = RaySetupOption.GET_OR_FAIL,
        ray_address: str = "auto",
        setup_cluster_servlet: ClusterServletSetupOption = ClusterServletSetupOption.GET_OR_CREATE,
        create_process_params: Optional["CreateProcessParams"] = None,
    ):
        import ray

        # The initialization of the obj_store needs to be in a separate method
        # so the HTTPServer actually initalizes the obj_store,
        # and it doesn't get created and destroyed when
        # caddy runs http_server.py as a module.

        # ClusterServlet essentially functions as a global state/metadata store
        # for all nodes connected to this Ray cluster.

        # If the servlet name is already set, the obj_store has already been initialized
        if self.servlet_name is not None:
            return

        from runhouse.resources.hardware.ray_utils import kill_actors

        # Only if ray is not initialized do we attempt a setup process.
        if not ray.is_initialized():
            if setup_ray == RaySetupOption.TEST_PROCESS:
                # When we run ray.init() with no address provided
                # and no Ray is running, it will start a new Ray cluster,
                # but one that is only exposed to this process. This allows us to
                # run unit tests without starting bare metal Ray clusters on each machine.
                ray.init(
                    ignore_reinit_error=True,
                    logging_level=logging.ERROR,
                    namespace="runhouse",
                )
            else:
                ray.init(
                    address=ray_address,
                    ignore_reinit_error=True,
                    logging_level=logging.ERROR,
                    namespace="runhouse",
                )

        # Now, we expect to be connected to an initialized Ray instance.
        if setup_cluster_servlet == ClusterServletSetupOption.FORCE_CREATE:
            kill_actors(namespace="runhouse", gracefully=False)

        create_if_not_exists = (
            setup_cluster_servlet != ClusterServletSetupOption.GET_OR_FAIL
        )
        self.cluster_servlet = get_cluster_servlet(
            create_if_not_exists=create_if_not_exists,
            runtime_env=create_process_params.runtime_env
            if create_process_params
            else None,
        )
        if self.cluster_servlet is None:
            # TODO: logger.<method> is not printing correctly here when doing `runhouse server start`.
            # Fix this and general logging.
            logger.warning(
                "Warning, cluster servlet is not initialized. Object Store operations will not work."
            )

        # There are 3 operating modes of the KV store:
        # servlet_name is set, has_local_storage is True: This is an Servlet with a local KV store.
        # servlet_name is set, has_local_storage is False: This is an ObjStore class that is not an Servlet,
        #   but wants to proxy its writes to a running Servlet.
        # servlet_name is unset, has_local_storage is False: This is an ObjStore class that by default only looks at
        #   the global KV store and other servlets.
        if not servlet_name and has_local_storage:
            raise ValueError(
                "Must provide a servlet name if the servlet has local storage."
            )

        # There can only be one initialized Servlet with a given name AND with local storage.
        if has_local_storage and servlet_name:
            if await self.ais_servlet_name_initialized(servlet_name):
                raise ValueError(
                    f"There already exists an Servlet with name {servlet_name}."
                )
            else:
                await self.aadd_servlet_initialized_args(
                    servlet_name, create_process_params
                )

        self.servlet_name = servlet_name
        self.has_local_storage = has_local_storage
        if self.has_local_storage:
            self._kv_store = {}

        # Store a local copy of the cluster_config here
        self.cluster_config = await self.aget_cluster_config(refresh=True)

        num_gpus = ray.cluster_resources().get("GPU", 0)
        cuda_visible_devices = list(range(int(num_gpus)))
        os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, cuda_visible_devices))

        # Add to the sys.path the path that you need to prepend
        paths_to_prepend = await self.aget_paths_to_prepend_in_new_processes()
        for path in paths_to_prepend:
            if path not in sys.path:
                sys.path.insert(0, path)

        # Set env vars that need to be set on creation
        env_vars_to_set = await self.aget_env_vars_to_set_in_new_processes()
        self.set_process_env_vars_local(env_vars_to_set)

        # Set env vars that were passed in initialization, these should override
        # any env vars that were set globally via the previous global set
        if create_process_params and create_process_params.env_vars:
            self.set_process_env_vars_local(create_process_params.env_vars)

    def initialize(
        self,
        servlet_name: Optional[str] = None,
        has_local_storage: bool = False,
        setup_ray: RaySetupOption = RaySetupOption.GET_OR_FAIL,
        ray_address: str = "auto",
        setup_cluster_servlet: ClusterServletSetupOption = ClusterServletSetupOption.GET_OR_CREATE,
        runtime_env: Optional[Dict] = None,
    ):
        return sync_function(self.ainitialize)(
            servlet_name,
            has_local_storage,
            setup_ray,
            ray_address,
            setup_cluster_servlet,
            runtime_env,
        )

    # Run bash locally, used by both NodeServlet and Servlet
    async def arun_with_logs_local(
        self, cmd: str, require_outputs: bool = True, run_name: Optional[str] = None
    ):
        log_ctx = None
        try:
            stream_logs = False
            if run_name is not None:
                self.active_bash_run_names.add(run_name)
                log_ctx = LogToFolder(name=run_name)
                log_ctx.__enter__()
                stream_logs = True

            res = await arun_in_thread(
                run_with_logs,
                cmd,
                stream_logs=stream_logs,
                require_outputs=require_outputs,
            )
            return res
        finally:
            if log_ctx is not None:
                log_ctx.__exit__(None, None, None)
                self.active_bash_run_names.remove(run_name)

    ###################################################
    # Node servlet functionality, mostly stored in cluster servlet for now
    ###################################################
    def node_servlet_name_for_ip(self, ip: str):
        # When testing and forwarding to local, the client side internal IP may be set as localhost,
        # and we need to convert it to the actual internal IP that the NodeServlet was initialized with.
        if ip == "localhost":
            ip = self.get_internal_ips()[0]
        return f"node_servlet_{ip}"

    async def ainitialize_node_servlets(self):
        """
        Initialize node servlets: This can't be done in the cluster servlet constructor because there'll be a circular
        dependency between the cluster servlet and the node servlets. We'll call this only on the head node in
        the server
        """
        import ray

        from runhouse.servers.node_servlet import NodeServlet

        names = []
        for ip in self.get_internal_ips():
            resources = {f"node:{ip}": 0.001}
            node_servlet_name = self.node_servlet_name_for_ip(ip)
            ray.remote(NodeServlet).options(
                name=node_servlet_name,
                get_if_exists=True,
                lifetime="detached",
                namespace="runhouse",
                max_concurrency=1000,
                resources=resources,
                num_cpus=0,
            ).remote()
            names.append(node_servlet_name)

        await self.acall_actor_method(
            self.cluster_servlet, "aset_node_servlet_names", names
        )

    async def aget_node_servlet_names(self):
        return await self.acall_actor_method(
            self.cluster_servlet, "aget_node_servlet_names"
        )

    def run_bash_command_on_node_or_process(
        self,
        command: str,
        require_outputs: bool = False,
        run_name: Optional[str] = None,
        node_ip: Optional[str] = None,
        process: Optional[str] = None,
    ):
        return sync_function(self.arun_bash_command_on_node_or_process)(
            command, require_outputs, run_name, node_ip, process
        )

    async def arun_bash_command_on_node_or_process(
        self,
        command: str,
        require_outputs: bool = False,
        run_name: Optional[str] = None,
        node_ip: Optional[str] = None,
        process: Optional[str] = None,
    ):
        import ray

        if node_ip and process:
            raise ValueError("Cannot specify both node_ip and process.")

        if not node_ip and not process:
            raise ValueError("Must specify either node_ip or process.")

        if node_ip:
            servlet_name = self.node_servlet_name_for_ip(node_ip)
        elif process:
            servlet_name = process

        actor_to_run_command_on = ray.get_actor(servlet_name, namespace="runhouse")
        return await self.acall_actor_method(
            actor_to_run_command_on,
            "arun_with_logs_local",
            command,
            require_outputs,
            run_name,
        )

    async def arun_bash_command_on_all_nodes(
        self, command: str, require_outputs: bool = False
    ):
        import ray

        node_servlet_names = await self.aget_node_servlet_names()
        node_servlet_actors = [
            ray.get_actor(name, namespace="runhouse") for name in node_servlet_names
        ]
        results = await asyncio.gather(
            *[
                self.acall_actor_method(
                    node_servlet, "arun_with_logs_local", command, require_outputs
                )
                for node_servlet in node_servlet_actors
            ]
        )
        return results

    def get_process_servlet(self):
        """
        If this is a servlet object store, then we are within a Runhouse process.
        Return the process so it can be used for Runhouse primitives.
        """
        if self.servlet_name is not None and self.has_local_storage:
            # Each process is stored within itself, I believe
            return self.get(self.servlet_name)

    def get_process_name(self) -> str:
        """
        If this is a servlet in the object store, then we are inside a Runhouse process.
        Return the process name.
        """
        return (
            self.servlet_name
            if (self.servlet_name is not None and self.has_local_storage)
            else None
        )

    def get_internal_ips(self):
        """Get list of internal IPs of all nodes in the cluster."""
        import ray

        cluster_config = self.get_cluster_config()
        if cluster_config.get("compute_properties", {}).get("internal_ips", []):
            return cluster_config.get("compute_properties").get("internal_ips")
        else:
            if not ray.is_initialized():
                raise ConnectionError("Ray is not initialized.")

            cluster_nodes = ray.nodes()
            return [node["NodeManagerAddress"] for node in cluster_nodes]

    ##############################################
    # Generic helpers
    ##############################################
    async def acall_servlet_method(
        self,
        servlet_name: str,
        method: str,
        *args,
        use_servlet_cache: bool = True,
        **kwargs,
    ):
        import ray

        servlet = self.get_servlet(servlet_name, use_servlet_cache=use_servlet_cache)
        if servlet is None:
            raise ObjStoreError(f"Got None servlet for servlet_name {servlet_name}.")
        try:
            return await ObjStore.acall_actor_method(servlet, method, *args, **kwargs)
        except (ray.exceptions.RayActorError, ray.exceptions.OutOfMemoryError) as e:
            if isinstance(
                e, ray.exceptions.OutOfMemoryError
            ) or "died unexpectedly before finishing this task" in str(e):
                await self.adelete_servlet_contents(servlet_name)

            raise e

    @staticmethod
    async def acall_actor_method(
        actor: "ray.actor.ActorHandle", method: str, *args, **kwargs
    ):
        if actor is None:
            raise ObjStoreError("Attempting to call an actor method on a None actor.")
        return await getattr(actor, method).remote(*args, **kwargs)

    @staticmethod
    def call_actor_method(actor: "ray.actor.ActorHandle", method: str, *args, **kwargs):
        import ray

        if actor is None:
            raise ObjStoreError("Attempting to call an actor method on a None actor.")

        return ray.get(getattr(actor, method).remote(*args, **kwargs))

    def get_servlet(
        self,
        name: str,
        create_process_params: Optional["CreateProcessParams"] = None,
        create: bool = False,
        raise_ex_if_not_found: bool = False,
        use_servlet_cache: bool = True,
        **kwargs,
    ):
        # Need to import these here to avoid circular imports
        import ray

        from runhouse.servers.servlet import Servlet

        if create_process_params is not None and (name != create_process_params.name):
            raise ValueError(
                "servlet name and create_process_params.name must be the same."
            )

        if use_servlet_cache and name in self.servlet_cache:
            return self.servlet_cache[name]

        # It may not have been cached, but does exist
        try:
            existing_actor = ray.get_actor(name, namespace="runhouse")
            if use_servlet_cache:
                self.servlet_cache[name] = existing_actor
            return existing_actor
        except ValueError:
            # ValueError: Failed to look up actor with name ...
            pass

        # Otherwise, create it
        if create:
            if not create_process_params:
                raise ValueError(
                    "Must provide create_process_params if creating a servlet."
                )

            ray_resources = copy.copy(create_process_params.compute)

            if ray_resources is None:
                # Put servlets on head node by default
                ray_resources = {"node_idx": 0}

            if "node_idx" in ray_resources and (
                "CPU" in ray_resources or "GPU" in ray_resources
            ):
                raise ValueError(
                    "Cannot specify both node_idx and CPU/GPU ray_resources for a process."
                )

            # Replace node_idx with actual node IP in Ray resources request
            node_idx = ray_resources.pop("node_idx", None)
            if node_idx is not None:
                cluster_ips = self.get_internal_ips()
                if node_idx >= len(cluster_ips):
                    raise ValueError(
                        f"Node index {node_idx} is out of bounds for cluster with {len(cluster_ips)} nodes."
                    )
                ray_resources[f"node:{cluster_ips[node_idx]}"] = 0.001

            # Check if requested resources are available
            available_resources = ray.available_resources()
            for k, v in ray_resources.items():
                if k not in available_resources or available_resources[k] < v:
                    raise Exception(
                        f"Requested resource {k}={v} is not available on the cluster. "
                        f"Available resources: {available_resources}"
                    )

            new_actor = (
                ray.remote(Servlet)
                .options(
                    name=name,
                    get_if_exists=True,
                    runtime_env=create_process_params.runtime_env,
                    # Default to 0 CPUs if not specified, Ray will default it to 1
                    num_cpus=ray_resources.pop("CPU", 0),
                    num_gpus=ray_resources.pop("GPU", None),
                    memory=ray_resources.pop("memory", None),
                    resources=ray_resources,
                    lifetime="detached",
                    namespace="runhouse",
                    max_concurrency=1000,
                )
                .remote(create_process_params=create_process_params)
            )

            # Make sure servlet is actually initialized
            _ = self.call_actor_method(new_actor, "keys_local")

            if use_servlet_cache:
                self.servlet_cache[name] = new_actor
            return new_actor

        else:
            if raise_ex_if_not_found:
                raise ObjStoreError(
                    f"Process {name} does not exist. Please send it to the cluster first."
                )
            else:
                return None

    @staticmethod
    def set_ctx(**ctx_args):
        from runhouse.servers.http.http_utils import RequestContext

        ctx = RequestContext(**ctx_args)
        return req_ctx.set(ctx)

    async def apopulate_ctx_locally(self):
        from runhouse.globals import configs

        token = configs.token
        return self.set_ctx(request_id=str(uuid.uuid4()), token=token)

    @staticmethod
    def unset_ctx(ctx_token):
        req_ctx.reset(ctx_token)

    ##############################################
    # Propagate den auth
    ##############################################
    # TODO: Maybe this function needs to be synchronous to propagate Den Auth changes immediately?
    # Guess we'll find out
    def is_den_auth_enabled(self):
        return (
            self.cluster_config.get("den_auth", False)
            if self.cluster_config is not None
            else False
        )

    async def aset_den_auth(self, den_auth: bool):
        await self.aset_cluster_config_value("den_auth", den_auth)

    async def aenable_den_auth(self):
        await self.aset_den_auth(True)

    async def adisable_den_auth(self):
        await self.aset_den_auth(False)

    def set_process_env_vars_local(self, env_vars: Dict[str, str]):
        set_env_vars_in_current_process(env_vars)

    async def aset_process_env_vars(self, servlet_name: str, env_vars: Dict[str, str]):
        if self.servlet_name == servlet_name and self.has_local_storage:
            self.set_process_env_vars_local(env_vars)
        else:
            return await self.acall_servlet_method(
                servlet_name, "aset_env_vars", env_vars
            )

    def set_process_env_vars(self, servlet_name: str, env_vars: Dict[str, str]):
        return sync_function(self.aset_process_env_vars)(servlet_name, env_vars)

    async def aset_env_vars_globally(self, env_vars: Dict[str, str]):
        return await self.acall_actor_method(
            self.cluster_servlet, "aset_env_vars_globally", env_vars
        )

    ##############################################
    # Cluster config state storage methods
    ##############################################
    async def aget_cluster_config(self, refresh: bool = False):
        if refresh or not self.cluster_config:
            if self.cluster_servlet is not None:
                self.cluster_config = await self.acall_actor_method(
                    self.cluster_servlet, "aget_cluster_config"
                )
            else:
                return {}

        return self.cluster_config

    def get_cluster_config(self):
        return sync_function(self.aget_cluster_config)()

    async def aset_cluster_config(self, config: Dict[str, Any]):
        self.cluster_config = await self.acall_actor_method(
            self.cluster_servlet, "aset_cluster_config", config
        )

    async def aset_cluster_config_value(self, key: str, value: Any):
        self.cluster_config = await self.acall_actor_method(
            self.cluster_servlet, "aset_cluster_config_value", key, value
        )

    def set_cluster_config_value(self, key: str, value: Any):
        sync_function(self.aset_cluster_config_value)(key, value)

    ##############################################
    # Auth cache internal functions
    ##############################################
    async def aresource_access_level(self, token: str, resource_uri: str):
        return await self.acall_actor_method(
            self.cluster_servlet,
            "aresource_access_level",
            token,
            resource_uri,
        )

    def resource_access_level(self, token: str, resource_uri: str):
        return sync_function(self.aresource_access_level)(token, resource_uri)

    async def aget_username(self, token: str):
        return await self.acall_actor_method(
            self.cluster_servlet, "aget_username", token
        )

    async def ahas_resource_access(self, token: str, resource_uri=None) -> bool:
        """Checks whether user has read or write access to a given module saved on the cluster."""
        from runhouse.globals import configs, rns_client
        from runhouse.rns.utils.api import ResourceAccess
        from runhouse.servers.http.http_utils import load_current_cluster_rns_address

        if token is None:
            # If no token is provided assume no access
            return False

        # The logged-in user always has full access to the cluster and its resources. This is especially
        # important if they flip on Den Auth without saving the cluster.

        # configs.token is the token stored on the cluster itself, which is itself a hashed subtoken
        config_token = configs.token
        if config_token:
            if config_token == token:
                return True

            if resource_uri and rns_client.validate_cluster_token(
                cluster_token=token, cluster_uri=resource_uri
            ):
                return True

        cluster_uri = load_current_cluster_rns_address()
        cluster_access = await self.aresource_access_level(token, cluster_uri)
        if cluster_access == ResourceAccess.WRITE:
            # if user has write access to cluster will have access to all resources
            return True

        if resource_uri != cluster_uri and cluster_access == ResourceAccess.READ:
            # If the user has READ access to the cluster and this isn't a cluster management
            # endpoint, they have access to all resources
            return True

        if resource_uri is None and cluster_access not in [
            ResourceAccess.WRITE,
            ResourceAccess.READ,
        ]:
            # If module does not have a name, must have access to the cluster
            return False

        resource_access_level = await self.aresource_access_level(token, resource_uri)
        return resource_access_level in [ResourceAccess.WRITE, ResourceAccess.READ]

    async def aclear_auth_cache(self, token: str = None):
        return await self.acall_actor_method(
            self.cluster_servlet, "aclear_auth_cache", token
        )

    ##############################################
    # Key to servlet where it is stored mapping
    ##############################################
    async def aadd_servlet_initialized_args(
        self, servlet_name: str, init_args: Optional["CreateProcessParams"] = None
    ):
        return await self.acall_actor_method(
            self.cluster_servlet,
            "aadd_servlet_initialized_args",
            servlet_name,
            init_args,
        )

    async def ais_servlet_name_initialized(self, servlet_name: str) -> bool:
        return await self.acall_actor_method(
            self.cluster_servlet, "ais_servlet_name_initialized", servlet_name
        )

    async def aget_all_initialized_servlet_args(
        self,
    ) -> Dict[str, "CreateProcessParams"]:
        return await self.acall_actor_method(
            self.cluster_servlet,
            "aget_all_initialized_servlet_args",
        )

    async def alist_processes(self):
        processes = await self.aget_all_initialized_servlet_args()
        return {k: v.model_dump() for k, v in processes.items()}

    def list_processes(self):
        return sync_function(self.alist_processes)()

    def get_all_initialized_servlet_args(self) -> Dict[str, "CreateProcessParams"]:
        return sync_function(self.aget_all_initialized_servlet_args)()

    async def aget_servlet_name_for_key(self, key: Any):
        return await self.acall_actor_method(
            self.cluster_servlet, "aget_servlet_name_for_key", key
        )

    def get_servlet_name_for_key(self, key: Any):
        return sync_function(self.aget_servlet_name_for_key)(key)

    async def _aput_servlet_name_for_key(self, key: Any, servlet_name: str):
        return await self.acall_actor_method(
            self.cluster_servlet, "aput_servlet_name_for_key", key, servlet_name
        )

    async def _apop_servlet_name_for_key(self, key: Any, *args) -> str:
        return await self.acall_actor_method(
            self.cluster_servlet, "apop_servlet_name_for_key", key, *args
        )

    async def aget_paths_to_prepend_in_new_processes(self) -> str:
        return await self.acall_actor_method(
            self.cluster_servlet, "aget_paths_to_prepend_in_new_processes"
        )

    async def aadd_path_to_prepend_in_new_processes(self, path: str):
        return await self.acall_actor_method(
            self.cluster_servlet, "aadd_path_to_prepend_in_new_processes", path
        )

    async def aget_env_vars_to_set_in_new_processes(self) -> Dict[str, str]:
        return await self.acall_actor_method(
            self.cluster_servlet, "aget_env_vars_to_set_in_new_processes"
        )

    ##############################################
    # Remove Servlet
    ##############################################
    async def aclear_all_references_to_servlet_name(self, servlet_name: str):
        return await self.acall_actor_method(
            self.cluster_servlet,
            "aclear_all_references_to_servlet_name",
            servlet_name,
        )

    ##############################################
    # KV Store: Keys
    ##############################################
    async def akeys_for_servlet_name(self, servlet_name: str) -> List[Any]:
        return await self.acall_servlet_method(servlet_name, "keys_local")

    def keys_for_servlet_name(self, servlet_name: str) -> List[Any]:
        return sync_function(self.akeys_for_servlet_name)(servlet_name)

    def keys_local(self) -> List[Any]:
        if self.has_local_storage:
            return list(self._kv_store.keys())
        else:
            return []

    async def akeys(self) -> List[Any]:
        # Return keys across the cluster, not only in this process
        return await self.acall_actor_method(
            self.cluster_servlet, "aget_key_to_servlet_name_dict_keys"
        )

    def keys(self) -> List[Any]:
        return sync_function(self.akeys)()

    ##############################################
    # KV Store: Put
    ##############################################
    async def aput_for_servlet_name(
        self,
        servlet_name: str,
        key: Any,
        data: Any,
        serialization: Optional[str] = None,
    ):
        return await self.acall_servlet_method(
            servlet_name,
            "aput_local",
            key,
            data=data,
            serialization=serialization,
        )

    async def aput_local(self, key: Any, value: Any):
        if self.has_local_storage:
            self._kv_store[key] = value
            await self._aput_servlet_name_for_key(key, self.servlet_name)
        else:
            raise NoLocalObjStoreError()

    async def aput(
        self,
        key: Any,
        value: Any,
        process: Optional[str] = None,
        serialization: Optional[str] = None,
        create_servlet_if_not_exists: bool = False,
    ):
        from runhouse.servers.http.http_utils import CreateProcessParams

        # Before replacing something else, check if this op will even be valid.
        if process is None and self.servlet_name is None:
            raise NoLocalObjStoreError()

        # If it was not specified, we want to put into our own servlet_name
        process = process or self.servlet_name

        if self.get_servlet(process) is None:
            if create_servlet_if_not_exists:
                self.get_servlet(
                    name=process,
                    create_process_params=CreateProcessParams(name=process),
                    create=True,
                )
            else:
                raise ObjStoreError(
                    f"Process {process} does not exist; cannot put key {key} there."
                )

        # If it does exist somewhere, no more!
        if await self.aget(key, default=None) is not None:
            logger.warning("Key already exists in some process, overwriting.")
            await self.apop(key)

        if self.has_local_storage and process == self.servlet_name:
            if serialization is not None:
                raise ObjStoreError(
                    "We should never reach this branch if serialization is not None."
                )
            await self.aput_local(key, value)
        else:
            await self.aput_for_servlet_name(process, key, value, serialization)

    def put(
        self,
        key: Any,
        value: Any,
        process: Optional[str] = None,
        serialization: Optional[str] = None,
        create_servlet_if_not_exists: bool = False,
    ):
        return sync_function(self.aput)(
            key, value, process, serialization, create_servlet_if_not_exists
        )

    ##############################################
    # KV Store: Get
    ##############################################
    async def aget_from_servlet_name(
        self,
        servlet_name: str,
        key: Any,
        default: Optional[Any] = None,
        serialization: Optional[str] = None,
        remote: bool = False,
    ):
        logger.info(f"Getting {key} from servlet {servlet_name}")
        return await self.acall_servlet_method(
            servlet_name,
            "aget_local",
            key,
            default=default,
            serialization=serialization,  # Crucial that this is a kwarg, or the wrapper doesn't pick it up!!
            remote=remote,
        )

    def get_from_servlet_name(
        self,
        servlet_name: str,
        key: Any,
        default: Optional[Any] = None,
        serialization: Optional[str] = None,
        remote: bool = False,
    ):
        return sync_function(self.aget_from_servlet_name)(
            servlet_name, key, default, serialization, remote
        )

    def get_local(self, key: Any, default: Optional[Any] = None, remote: bool = False):
        if self.has_local_storage:
            try:
                res = self._kv_store[key]
                if remote:
                    if hasattr(res, "config"):
                        return res.config()
                    else:
                        raise ValueError(
                            f"Cannot return remote for non-Resource object of type {type(res)}."
                        )
                return res
            except KeyError as e:
                if default == KeyError:
                    raise e
                return default
        else:
            if default == KeyError:
                raise KeyError(f"No local store exists; key {key} not found.")
            return default

    async def aget(
        self,
        key: Any,
        serialization: Optional[str] = None,
        remote: bool = False,
        default: Optional[Any] = None,
    ):
        servlet_name_containing_key = await self.aget_servlet_name_for_key(key)

        if not servlet_name_containing_key:
            if default == KeyError:
                raise KeyError(f"No local store exists; key {key} not found.")
            return default

        if servlet_name_containing_key == self.servlet_name and self.has_local_storage:
            # Short-circuit route if we're already in the right process
            res = self.get_local(
                key,
                remote=remote,
                default=default,
            )
        else:
            # Note, if serialization is not None here and remote is True we won't enter the block below,
            # because the Servlet already packaged the config into a Response object. This is desired, as we
            # only want the remote object to be reconstructed when it's being returned to the user, which would
            # not be here if serialization is not None (probably the HTTPClient).
            res = await self.aget_from_servlet_name(
                servlet_name_containing_key,
                key,
                default=default,
                serialization=serialization,
                remote=remote,
            )

        # When the user called the obj_store.get with remote directly, we need to
        # package the config back into the remote object here before returning it.
        if (
            remote
            and (serialization is None or serialization == "none")
            and isinstance(res, dict)
            and "resource_type" in res
        ):
            config = res
            if config.get("system") == await self.aget_cluster_config():
                from runhouse import here

                config["system"] = here
            from runhouse.resources.resource import Resource

            res_copy = Resource.from_config(config=config, dryrun=True)
            return res_copy

        return res

    def get(
        self,
        key: Any,
        serialization: Optional[str] = None,
        remote: bool = False,
        default: Optional[Any] = None,
    ):
        return sync_function(self.aget)(key, serialization, remote, default)

    ##############################################
    # KV Store: Contains
    ##############################################
    async def acontains_for_servlet_name(self, servlet_name: str, key: Any):
        return await self.acall_servlet_method(servlet_name, "acontains_local", key)

    def contains_local(self, key: Any):
        if self.has_local_storage:
            return key in self._kv_store
        else:
            return False

    async def acontains(self, key: Any):
        if self.contains_local(key):
            return True

        servlet_name = await self.aget_servlet_name_for_key(key)
        if servlet_name == self.servlet_name and self.has_local_storage:
            raise ObjStoreError(
                "Key not found in kv store despite servlet specifying that it is here."
            )

        if servlet_name is None:
            return False

        return await self.acontains_for_servlet_name(servlet_name, key)

    def contains(self, key: Any):
        return sync_function(self.acontains)(key)

    ##############################################
    # KV Store: Pop
    ##############################################
    async def apop_from_servlet_name(
        self,
        servlet_name: str,
        key: Any,
        serialization: Optional[str] = "pickle",
        *args,
    ) -> Any:
        return await self.acall_servlet_method(
            servlet_name,
            "apop_local",
            key,
            serialization,
            *args,
        )

    async def apop_local(self, key: Any, *args) -> Any:
        if self.has_local_storage:
            try:
                res = self._kv_store.pop(key)
            except KeyError as key_err:
                # Return the default if it was provided, else raise the error as expected
                if args:
                    return args[0]
                else:
                    raise key_err

            # If the key was found in this process, we also need to pop it
            # from the global process for key cache.
            process = await self._apop_servlet_name_for_key(key, None)
            if process and process != self.servlet_name:
                raise ObjStoreError(
                    "The key was popped from this process, but the global process for key cache says it's in another one."
                )

            return res
        else:
            if args:
                return args[0]
            else:
                raise KeyError(f"No local store exists; key {key} not found.")

    async def apop(
        self, key: Any, serialization: Optional[str] = "pickle", *args
    ) -> Any:
        try:
            return await self.apop_local(key)
        except KeyError as e:
            key_err = e

        # The key was not found in this process
        # So, we check the global key to process cache to see if it's elsewhere
        servlet_name = await self.aget_servlet_name_for_key(key)
        if servlet_name:
            if servlet_name == self.servlet_name and self.has_local_storage:
                raise ObjStoreError(
                    "The key was not found in this process, but the global process for key cache says it's here."
                )
            else:
                # The key was found in another process, so we need to pop it from there
                return await self.apop_from_servlet_name(
                    servlet_name, key, serialization
                )
        else:
            # Was not found in any process
            if args:
                return args[0]
            else:
                raise key_err

    def pop(self, key: Any, serialization: Optional[str] = "pickle", *args) -> Any:
        return sync_function(self.apop)(key, serialization, *args)

    ##############################################
    # KV Store: Delete
    ##############################################
    async def adelete_for_servlet_name(self, servlet_name: str, key: Any):
        return await self.acall_servlet_method(servlet_name, "adelete_local", key)

    async def adelete_local(self, key: Any):
        await self.apop_local(key)

    async def adelete_servlet_contents(self, servlet_name: Any):
        import ray

        # delete the servlet actor and remove its references
        if servlet_name in self.servlet_cache:
            actor = self.servlet_cache[servlet_name]
            ray.kill(actor)
            del self.servlet_cache[servlet_name]
        deleted_keys = await self.aclear_all_references_to_servlet_name(servlet_name)
        return deleted_keys

    def delete_servlet_contents(self, servlet_name: Any):
        return sync_function(self.adelete_servlet_contents)(servlet_name)

    async def adelete_servlet_from_cache(self, servlet_name: Any):
        if servlet_name in self.servlet_cache:
            del self.servlet_cache[servlet_name]

    async def adelete(self, key: Union[Any, List[Any]]):
        keys_to_delete = [key] if isinstance(key, str) else key
        deleted_keys = []

        for key_to_delete in keys_to_delete:
            if key_to_delete in await self.aget_all_initialized_servlet_args():
                deleted_keys += await self.adelete_servlet_contents(key_to_delete)

            if key_to_delete in deleted_keys:
                continue

            if self.contains_local(key_to_delete):
                await self.adelete_local(key_to_delete)
                deleted_keys.append(key_to_delete)
            else:
                servlet_name = await self.aget_servlet_name_for_key(key_to_delete)
                if servlet_name == self.servlet_name and self.has_local_storage:
                    raise ObjStoreError(
                        "Key not found in kv store despite servlet specifying that it is here."
                    )
                if servlet_name is None:
                    raise KeyError(f"Key {key} not found in any process.")

                await self.adelete_for_servlet_name(servlet_name, key_to_delete)
                deleted_keys.append(key_to_delete)

    def delete(self, key: Union[Any, List[Any]]):
        return sync_function(self.adelete)(key)

    ##############################################
    # KV Store: Clear
    ##############################################
    async def aclear_for_servlet_name(self, servlet_name: str):
        return await self.acall_servlet_method(servlet_name, "aclear_local")

    async def aclear_local(self):
        if self.has_local_storage:
            # Use asyncio gather to run all the deletes concurrently
            await asyncio.gather(
                *[self.apop_local(k) for k in list(self._kv_store.keys())]
            )

    async def aclear(self):
        logger.warning("Clearing all keys from all servlets in the object store!")
        for servlet_name in await self.aget_all_initialized_servlet_args():
            if servlet_name == self.servlet_name and self.has_local_storage:
                await self.aclear_local()
            else:
                await self.aclear_for_servlet_name(servlet_name)

    def clear(self):
        return sync_function(self.aclear)()

    ##############################################
    # KV Store: Rename
    ##############################################
    async def arename_for_servlet_name(
        self, servlet_name: str, old_key: Any, new_key: Any
    ):
        return await self.acall_servlet_method(
            servlet_name,
            "arename_local",
            old_key,
            new_key,
        )

    async def arename_local(self, old_key: Any, new_key: Any):
        if self.servlet_name is None or not self.has_local_storage:
            raise NoLocalObjStoreError()

        obj = await self.apop(old_key)
        if obj is not None and hasattr(obj, "rns_address"):
            # Note - we set the obj.name here so the new_key is correctly turned into an rns_address, whether its
            # a full address or just a name. Then, the new_key is set to just the name so its store properly in the
            # kv store.
            obj.name = new_key  # new_key can be an rns_address! e.g. if called by Module.rename
            new_key = obj.name  # new_key is now just the name

        # By passing default, we don't throw an error if the key is not found
        await self.aput(new_key, obj, process=self.servlet_name)

    async def arename(self, old_key: Any, new_key: Any):
        # We also need to rename the resource itself
        servlet_name_containing_old_key = await self.aget_servlet_name_for_key(old_key)
        if (
            servlet_name_containing_old_key == self.servlet_name
            and self.has_local_storage
        ):
            await self.arename_local(old_key, new_key)
        else:
            await self.arename_for_servlet_name(
                servlet_name_containing_old_key, old_key, new_key
            )

    def rename(self, old_key: Any, new_key: Any):
        return sync_function(self.arename)(old_key, new_key)

    ##############################################
    # KV Store: Call
    ##############################################
    async def acall_for_servlet_name(
        self,
        servlet_name: str,
        key: Any,
        method_name: str,
        data: Any = None,
        serialization: Optional[str] = None,
        run_name: Optional[str] = None,
        stream_logs: bool = False,
        remote: bool = False,
    ):
        return await self.acall_servlet_method(
            servlet_name,
            "acall_local",
            key,
            method_name=method_name,
            data=data,
            serialization=serialization,
            run_name=run_name,
            stream_logs=stream_logs,
            remote=remote,
            ctx=dict(req_ctx.get()),
        )

    async def acall_local(
        self,
        key: str,
        method_name: Optional[str] = None,
        *args,
        run_name: Optional[str] = None,
        stream_logs: bool = False,
        remote: bool = False,
        **kwargs,
    ):
        """Base call functionality: Load the module, and call a method on it with args and kwargs. Nothing else.

        Handles calls on properties, methods, coroutines, and generators.

        """
        if self.servlet_name is None or not self.has_local_storage:
            raise NoLocalObjStoreError()

        log_ctx = None
        if stream_logs and run_name is not None:
            log_ctx = LogToFolder(name=run_name)
            log_ctx.__enter__()

        # Use a finally to track the active functions so that it is always removed
        request_id = req_ctx.get().request_id

        # There can be many function calls in one request_id, since request_id is tied to a call from the
        # client to the server.
        # We store with this func_call_id so we can easily pop the active call info out after the function
        # concludes. In theory we could use a tuple of (key, start_time, etc), but it doesn't accomplish much
        func_call_id = uuid.uuid4()
        self.active_function_calls[func_call_id] = ActiveFunctionCallInfo(
            key=key,
            method_name=method_name,
            request_id=request_id,
            start_time=time.time(),
            run_name=run_name,
        )
        try:
            res = await self._acall_local_helper(
                key,
                method_name,
                *args,
                run_name=run_name,
                stream_logs=stream_logs,
                remote=remote,
                **kwargs,
            )
        finally:
            del self.active_function_calls[func_call_id]
            if log_ctx:
                log_ctx.__exit__(None, None, None)

        return res

    async def _acall_local_helper(
        self,
        key: str,
        method_name: Optional[str] = None,
        *args,
        run_name: Optional[str] = None,
        stream_logs: bool = False,
        remote: bool = False,
        **kwargs,
    ):
        """acall_local primarily sets up the logging and tracking for the function call, then calls
        _acall_local_helper to actually do the work. This is so we can have a finally block in acall_local to clean up
        the active function calls tracking."""
        obj = self.get_local(key, default=KeyError)

        from runhouse.resources.module import Module
        from runhouse.resources.resource import Resource

        if self.is_den_auth_enabled():
            if not isinstance(obj, Resource) or obj.visibility not in [
                ResourceVisibility.UNLISTED,
            ]:
                ctx = req_ctx.get()
                if not ctx or not ctx.token:
                    raise PermissionError(
                        "No Runhouse token provided. Try running `$ runhouse login` or visiting "
                        "https://run.house/login to retrieve a token. If calling via HTTP, please "
                        "provide a valid token in the Authorization header.",
                    )

                # Setting to None in the case of non-resource or no rns_address will force auth to only
                # succeed if the user has WRITE or READ access to the cluster
                resource_uri = obj.rns_address if hasattr(obj, "rns_address") else None
                if key != self.servlet_name and not await self.ahas_resource_access(
                    ctx.token, resource_uri
                ):
                    # Do not validate access to the default process
                    raise PermissionError(
                        f"Unauthorized access to resource {key}.",
                    )

        # Process any inputs which need to be resolved
        args = [
            arg.fetch() if (isinstance(arg, Module) and arg._resolve) else arg
            for arg in args
        ]
        kwargs = {
            k: v.fetch() if (isinstance(v, Module) and v._resolve) else v
            for k, v in kwargs.items()
        }

        method_name = method_name or "__call__"

        method_is_coroutine = False
        try:
            if isinstance(obj, Module):
                # Force this to be fully local for Modules so we don't have any circular stuff calling into other
                # processes or systems.
                method = getattr(obj.local, method_name)
                module_signature = obj.signature(rich=True)
                method_is_coroutine = (
                    method_name in module_signature
                    and module_signature[method_name]["async"]
                    and not module_signature[method_name]["gen"]
                )
            else:
                method = getattr(obj, method_name)
                method_is_coroutine = inspect.iscoroutinefunction(method)
        except AttributeError:
            logger.debug(obj.__dict__)
            raise ValueError(f"Method {method_name} not found on module {obj}")

        if method_name in ["remote_anext", "remote_await"]:
            # If the method is a coroutine or generator, we need to call it with await
            # This is a special case because when the method comes through first time around, we notice that the
            # result is a coroutine, repackage it in a "FutureModule" or "AsyncGeneratorModule", and then return
            # it to the user. Then, when the user actually awaits, they call our magic `__await__` method, which
            # calls our synchronous dunder method, which is then caught here and *actually* awaited.
            logger.info(
                f"{self.servlet_name} process: Calling special method via await {method_name} on module {key}"
            )
            res = await method(*args, **kwargs)

        elif hasattr(method, "__call__") or method_name == "__call__":
            # If method is callable, call it and return the result
            logger.info(
                f"{self.servlet_name} process: Calling method {method_name} on module {key}"
            )

            res = await arun_in_thread(method, *args, **kwargs)

            # If this was a coroutine function (not a function returning a corotuine), we need to await it here,
            # and not use the FutureModule

            # Note that for async ops, we are using the single thread within our object store for the actual
            # execution of the async function. This means if it is poorly written user code that is blocking,
            # it will block the entire object store.
            if method_is_coroutine:
                res = await res

        else:
            if args and len(args) == 1:
                # if there's an arg, this is a "set" call on the property
                logger.info(
                    f"{self.servlet_name} servlet: Setting property {method_name} on module {key}"
                )
                if isinstance(obj, Module):
                    setattr(obj.local, method_name, args[0])
                else:
                    setattr(obj, method_name, args[0])
                res = None
            else:
                # Method is a property, return the value
                logger.info(
                    f"{self.servlet_name} servlet: Getting property {method_name} on module {key}"
                )
                res = method

        laziness_type = (
            "coroutine"
            if inspect.iscoroutine(res)
            else "generator"
            if inspect.isgenerator(res)
            else "async generator"
            if inspect.isasyncgen(res)
            else None
        )

        if laziness_type:
            # If the result is a coroutine or generator, we can't return it over the process boundary
            # and need to store it to be retrieved later. In this case we return a "retrievable".
            logger.debug(
                f"{self.servlet_name} servlet: Method {method_name} on module {key} is a {laziness_type}. "
                f"Storing result to be retrieved later at result key {res}."
            )
            fut = self._construct_call_retrievable(res, run_name, laziness_type)
            await self.aput_local(run_name, fut)
            return fut

        from runhouse.resources.resource import Resource

        if isinstance(res, Resource):
            if run_name and "--" not in run_name:
                # This is a user-specified name, so we want to override the existing name with it
                # and save the resource
                res.name = run_name or res.name
                await self.aput_local(res.name, res)

            if remote:
                # If we've reached this block then we know "--" is in run_name and it's an auto-generated name,
                # so we don't want override the existing name with it (as we do above with user-specified name)
                res.name = res.name or run_name

                # Need to save the resource in case we haven't yet (e.g. if run_name was auto-generated)
                await self.aput_local(res.name, res)
                # If remote is True and the result is a resource, we return just the config
                res = res.config()

        return res

    @staticmethod
    def _construct_call_retrievable(res, res_key, laziness_type):
        if laziness_type == "coroutine":
            from runhouse.resources.future_module import FutureModule

            # TODO make this one-time-use
            return FutureModule(future=res, name=res_key)

        elif laziness_type == "generator":
            from runhouse.resources.future_module import GeneratorModule

            return GeneratorModule(future=res, name=res_key)

        elif laziness_type == "async generator":
            from runhouse.resources.future_module import AsyncGeneratorModule

            return AsyncGeneratorModule(future=res, name=res_key)

        else:
            raise ValueError(f"Invalid laziness type {laziness_type}")

    @context_wrapper
    async def acall(
        self,
        key: str,
        method_name: Optional[str] = None,
        data: Any = None,
        serialization: Optional[str] = None,
        run_name: Optional[str] = None,
        stream_logs: bool = False,
        remote: bool = False,
    ):
        servlet_name_containing_key = await self.aget_servlet_name_for_key(key)
        if not servlet_name_containing_key:
            raise ObjStoreError(
                f"Key {key} not found in any process, cannot call method {method_name} on it."
            )

        # If there is no run_name, this call was made directly through the obj_store and we need to set one
        run_name = run_name or generate_default_name(
            prefix=key if method_name == "__call__" else f"{key}_{method_name}",
            precision="ms",  # Higher precision because we see collisions within the same second
            sep="--",
        )

        if servlet_name_containing_key == self.servlet_name and self.has_local_storage:
            from runhouse.servers.http.http_utils import deserialize_data

            deserialized_data = deserialize_data(data, serialization) or {}
            args, kwargs = deserialized_data.get("args", []), deserialized_data.get(
                "kwargs", {}
            )

            res = await self.acall_local(
                key,
                method_name,
                run_name=run_name,
                stream_logs=stream_logs,
                remote=remote,
                *args,
                **kwargs,
            )
        else:
            # In the background, retrieve logs for this run_name with self.alogs_for_servlet_name
            # with the run_name and print them while the call runs
            logs_task = None
            if stream_logs:

                async def print_logs():
                    async for logs in self.alogs_for_run_name(
                        run_name, servlet_name=servlet_name_containing_key
                    ):
                        for log in logs:
                            print(f"({key}): {log}", end="")

                logs_task = asyncio.create_task(print_logs())

            res = await self.acall_for_servlet_name(
                servlet_name_containing_key,
                key,
                method_name,
                data=data,
                serialization=serialization,
                run_name=run_name,
                stream_logs=stream_logs,
                remote=remote,
            )

            # TODO we only do this to prevent the task from being garbage collected after
            # the function completes. We should put it somewhere else so it doesn't need to
            # block execution for every internal call.
            if stream_logs:
                await logs_task

        if remote and isinstance(res, dict) and "resource_type" in res:
            config = res
            # Config is condensed by default, so system may just be a string
            if isinstance(config.get("system"), dict):
                system_name = config.get("system").get("name")
            else:
                system_name = config.get("system")
            if system_name == (await self.aget_cluster_config()).get("name"):
                from runhouse import here

                config["system"] = here
            from runhouse.resources.resource import Resource

            res_copy = Resource.from_config(config=config, dryrun=True)
            return res_copy

        return res

    def call(
        self,
        key: str,
        method_name: Optional[str] = None,
        data: Any = None,
        serialization: Optional[str] = None,
        run_name: Optional[str] = None,
        stream_logs: bool = False,
        remote: bool = False,
    ):
        return sync_function(self.acall)(
            key,
            method_name,
            data,
            serialization,
            run_name,
            stream_logs,
            remote,
        )

    async def alogs_for_run_name(
        self,
        run_name: str,
        servlet_name: Optional[str] = None,
        key: Optional[str] = None,
        node_ip: Optional[str] = None,
    ):
        if not servlet_name and not key and not node_ip:
            raise ValueError("Must provide either servlet_name or key.")

        if (servlet_name and node_ip) or (servlet_name and key) or (key and node_ip):
            raise ValueError("Must provide only one of servlet_name, key, or node_ip.")

        if node_ip:
            servlet_name = self.node_servlet_name_for_ip(node_ip)
        if key:
            servlet_name = await self.aget_servlet_name_for_key(key)

        servlet = self.get_servlet(servlet_name)
        async for log_ref in servlet.alogs_local.remote(run_name=run_name):
            logs = await log_ref
            yield logs

    @staticmethod
    def _get_logfiles(log_key, log_type=None):
        if not log_key:
            return None
        key_logs_path = Path(RH_LOGFILE_PATH) / log_key
        if key_logs_path.exists():
            # Logs are like: `.rh/logs/key/key.[out|err]`
            glob_pattern = (
                "*.out"
                if log_type == "stdout"
                else "*.err"
                if log_type == "stderr"
                else "*.[oe][ur][tr]"
            )
            return [str(f.absolute()) for f in key_logs_path.glob(glob_pattern)]
        else:
            return None

    @staticmethod
    def open_new_logfiles(key, open_files):
        logfiles = ObjStore._get_logfiles(key)
        if logfiles:
            for f in logfiles:
                if f not in [o.name for o in open_files]:
                    logger.info(f"Streaming logs from {f}")
                    open_files.append(open(f, "r"))
        return open_files

    async def alogs_local(self, run_name: str, bash_run: bool = False):
        def function_call_complete():
            active_run_names = [v.run_name for v in self.active_function_calls.values()]
            return run_name not in active_run_names

        def bash_run_complete():
            return run_name not in self.active_bash_run_names

        logging_complete = bash_run_complete if bash_run else function_call_complete
        async for lines in self._alogs_for_run_name_on_disk(logging_complete, run_name):
            yield lines

    async def _alogs_for_run_name_on_disk(
        self, logging_complete_fn: Callable, run_name: str
    ):
        logger.info(f"Begin streaming logs for run_name: '{run_name}'")
        open_logfiles = []

        # Wait for a maximum of 5 seconds for the log files to be created
        sleeps = 0
        while (sleeps * LOGS_TO_SHOW_UP_CHECK_TIME) <= MAX_LOGS_TO_SHOW_UP_WAIT_TIME:
            open_logfiles = ObjStore.open_new_logfiles(run_name, open_logfiles)
            if open_logfiles:
                break
            else:
                await asyncio.sleep(LOGS_TO_SHOW_UP_CHECK_TIME)
                sleeps += 1

        if not open_logfiles:
            logger.warning(f"No logfiles found for call {run_name}")

        call_in_progress = True

        logger.info(f"Found logfiles for key '{run_name}', beginning streaming.")
        try:
            while call_in_progress:
                # If the call is not in the active calls, it has finished (even if it finished before we
                # started streaming logs)
                if logging_complete_fn():
                    call_in_progress = False
                else:
                    await asyncio.sleep(LOGGING_WAIT_TIME)
                # Grab all the lines written to all the log files since the last time we checked, including
                # any new log files that have been created
                open_logfiles = ObjStore.open_new_logfiles(run_name, open_logfiles)
                ret_lines = []
                for i, f in enumerate(open_logfiles):
                    file_lines = f.readlines()
                    if file_lines:
                        ret_lines += file_lines
                if ret_lines:
                    logger.debug(f"Yielding logs for key {run_name}")
                    yield ret_lines
        finally:
            for f in open_logfiles:
                f.close()

    ##############################################
    # Get several keys for function initialization utilities
    ##############################################
    async def aget_list(self, keys: List[str], default: Optional[Any] = None):
        return await asyncio.gather(
            *[self.aget(key, default=default or key) for key in keys]
        )

    def get_list(self, keys: List[str], default: Optional[Any] = None):
        return sync_function(self.aget_list)(keys, default)

    async def aget_obj_refs_list(self, keys: List[Any]):

        return await asyncio.gather(
            *[
                self.aget(key, default=key) if isinstance(key, str) else key
                for key in keys
            ]
        )

    def get_obj_refs_list(self, keys: List[Any]):
        return sync_function(self.aget_obj_refs_list)(keys)

    async def aget_obj_refs_dict(self, d: Dict[Any, Any]):
        return {
            k: await self.aget(v, default=v) if isinstance(v, str) else v
            for k, v in d.items()
        }

    def get_obj_refs_dict(self, d: Dict[Any, Any]):
        return sync_function(self.aget_obj_refs_dict)(d)

    ##############################################
    # More specific helpers
    ##############################################
    async def aput_resource(
        self,
        serialized_data: Any,
        serialization: Optional[str] = None,
        process: Optional[str] = None,
    ) -> "Response":
        from runhouse.servers.http.http_utils import deserialize_data

        if process is None and self.servlet_name is None:
            raise ObjStoreError("No process name provided and no servlet name set.")

        process = process or self.servlet_name
        if self.has_local_storage and process == self.servlet_name:
            resource_config, state, dryrun = tuple(
                deserialize_data(serialized_data, serialization)
            )
            return await self.aput_resource_local(resource_config, state, dryrun)

        # Normally, serialization and deserialization happens within the servlet
        # However, if we're putting a process, we need to deserialize it here and
        # actually create the corresponding env servlet.
        resource_config, _, _ = tuple(deserialize_data(serialized_data, serialization))

        return await self.acall_servlet_method(
            process,
            "aput_resource_local",
            data=serialized_data,
            serialization=serialization,
        )

    def put_resource(
        self,
        serialized_data: Any,
        serialization: Optional[str] = None,
        process: Optional[str] = None,
    ) -> "Response":
        return sync_function(self.aput_resource)(
            serialized_data, serialization, process
        )

    async def aput_resource_local(
        self,
        resource_config: Dict[str, Any],
        state: Dict[Any, Any],
        dryrun: bool,
    ) -> str:
        from runhouse.resources.module import Module
        from runhouse.resources.resource import Resource

        state = state or {}
        # Resolve any sub-resources which are string references to resources already sent to this cluster.
        # We need to pop the resource's own name so it doesn't get resolved if it's already present in the
        # obj_store.
        name = resource_config.pop("name")
        subtype = resource_config.pop("resource_subtype")
        provider = (
            resource_config.pop("provider") if "provider" in resource_config else None
        )

        resource_config = await self.aget_obj_refs_dict(resource_config)
        resource_config["name"] = name
        resource_config["resource_subtype"] = subtype
        if provider:
            resource_config["provider"] = provider

        logger.debug(f"Message received from client to construct resource: {name}")

        resource = Resource.from_config(config=resource_config, dryrun=dryrun)

        for attr, val in state.items():
            setattr(resource, attr, val)

        name = resource.name or generate_default_name(prefix=resource.RESOURCE_TYPE)
        if isinstance(resource, Module):
            resource.rename(name)
        else:
            resource.name = name

        await self.aput_local(resource.name, resource)

        # Return the name in case we had to set it
        return resource.name

    ##############################################
    # Cluster info methods
    ##############################################
    def keys_with_info(self):
        if not self.has_local_storage or self.servlet_name is None:
            raise NoLocalObjStoreError()

        # Need to copy to avoid race conditions
        current_active_function_calls = copy.copy(self.active_function_calls)

        keys_and_info = {}
        for k, v in self._kv_store.items():
            cls = type(v)
            py_module = cls.__module__
            cls_name = (
                cls.__qualname__
                if py_module == "builtins"
                else (py_module + "." + cls.__qualname__)
            )

            active_fn_calls = [
                call_info.model_dump()
                for call_info in current_active_function_calls.values()
                if call_info.key == k
            ]

            keys_and_info[k] = {
                "resource_type": cls_name,
                "active_function_calls": active_fn_calls,
            }

        return keys_and_info

    async def astatus(self, send_to_den: bool = False):
        return await self.acall_actor_method(
            self.cluster_servlet, "status", send_to_den
        )

    def status(self, send_to_den: bool = False):
        return sync_function(self.astatus)(send_to_den=send_to_den)

    ##############################################
    # Globally modify the sys.path in all processes
    ##############################################
    async def aadd_sys_path_to_all_processes(self, path: str):
        tasks = [
            self.acall_servlet_method(servlet_name, "aprepend_to_sys_path", path)
            for servlet_name in await self.aget_all_initialized_servlet_args()
        ]
        await asyncio.gather(*tasks)

    def add_sys_path_to_all_processes(self, path: str):
        return sync_function(self.aadd_sys_path_to_all_processes)(path)

    ##############################################
    # Interact across nodes
    ##############################################
    async def ainstall_package_in_all_nodes_and_processes(
        self,
        package: "Package",
        conda_env_name: Optional[str] = None,
        force_sync_local: bool = False,
    ):
        from runhouse.resources.packages import InstallTarget, Package
        from runhouse.resources.packages.package import INSTALL_METHODS

        # Step 1: Make sure that the package is a Package object and the path that it may reference exists if there's path involved
        if not isinstance(package, Package):
            raise ValueError("The package must be a Package object.")

        if isinstance(package.install_target, InstallTarget) and not os.path.exists(
            package.install_target.full_local_path_str()
        ):
            raise FileNotFoundError(
                f"Path {package.install_target.full_local_path_str()} does not exist."
            )

        logger.info(f"Installing {str(package)} with method {package.install_method}.")

        if package.install_method == "pip":

            # If this is a generic pip package, with no version pinned, we want to check if there is a version
            # already installed. If there is, then we ignore preferred version and leave the existing version.
            # The user can always force a version install by doing `numpy==2.0.0` for example. Else, we install
            # the preferred version, that matches their local.
            if (
                is_python_package_string(package.install_target)
                and package.preferred_version is not None
            ):
                # Check if this is installed
                retcode = run_with_logs(
                    f"python -c \"import importlib.util; exit(0) if importlib.util.find_spec('{package.install_target}') else exit(1)\"",
                )
                if retcode != 0 or force_sync_local:
                    package.install_target = (
                        f"{package.install_target}=={package.preferred_version}"
                    )

            install_cmd = package._pip_install_cmd(conda_env_name=conda_env_name)
            logger.info(f"Running via install_method pip: {install_cmd}")
            run_cmd_results = await self.arun_bash_command_on_all_nodes(install_cmd)
            if any(run_cmd_result != 0 for run_cmd_result in run_cmd_results):
                raise RuntimeError(
                    f"Pip install {install_cmd} failed, check that the package exists and is available for your platform."
                )

        elif package.install_method == "conda":
            install_cmd = package._conda_install_cmd(conda_env_name=conda_env_name)
            logger.info(f"Running via install_method conda: {install_cmd}")
            run_cmd_results = await self.arun_bash_command_on_all_nodes(install_cmd)
            if any(run_cmd_result != 0 for run_cmd_result in run_cmd_results):
                raise RuntimeError(
                    f"Conda install {install_cmd} failed, check that the package exists and is "
                    "available for your platform."
                )

        elif package.install_method == "reqs":
            install_cmd = package._reqs_install_cmd(conda_env_name=conda_env_name)
            if install_cmd:
                logger.info(f"Running via install_method reqs: {install_cmd}")
                run_cmd_results = await self.arun_bash_command_on_all_nodes(install_cmd)
                if any(run_cmd_result != 0 for run_cmd_result in run_cmd_results):
                    raise RuntimeError(
                        f"Reqs install {install_cmd} failed, check that the package exists and is available for your platform."
                    )
            else:
                logger.info(
                    f"{package.install_target.full_local_path_str()}/requirements.txt not found, skipping reqs install"
                )

        else:
            if package.install_method != "local":
                raise ValueError(
                    f"Unknown install method {package.install_method}. Must be one of {INSTALL_METHODS}"
                )

        # Need to append to path
        if package.install_method in ["local", "reqs"]:
            if isinstance(package.install_target, InstallTarget):
                await self.aadd_sys_path_to_all_processes(
                    package.install_target.full_local_path_str()
                )

                await self.aadd_path_to_prepend_in_new_processes(
                    package.install_target.full_local_path_str()
                )



import copy
import os
from pathlib import Path

from typing import Dict

import yaml

from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret
from runhouse.resources.secrets.utils import _check_file_for_mismatches


class GitHubSecret(ProviderSecret):
    """
    .. note::
            To create a GitHubSecret, please use the factory method :func:`provider_secret` with ``provider="github"``.
    """

    # values format: {"oauth_token": oath_token}
    _PROVIDER = "github"
    _DEFAULT_CREDENTIALS_PATH = "~/.config/gh/hosts.yml"

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return GitHubSecret(**config, dryrun=dryrun)

    def _write_to_file(
        self,
        path: str,
        values: Dict = None,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        new_secret = copy.deepcopy(self)
        if not _check_file_for_mismatches(
            path, self._from_path(path), values, overwrite
        ):
            config = {}

            full_path = os.path.expanduser(path)
            if Path(full_path).exists():
                with open(full_path, "r") as stream:
                    config = yaml.safe_load(stream)
            config["github.com"] = values

            Path(full_path).parent.mkdir(parents=True, exist_ok=True)
            with open(full_path, "w") as yaml_file:
                yaml.dump(config, yaml_file, default_flow_style=False)

            if write_config:
                new_secret._add_to_rh_config(path)

        new_secret._values = None
        new_secret.path = path
        return new_secret

    def _from_path(self, path: str = None):
        config = {}
        if path and os.path.exists(os.path.expanduser(path)):
            with open(os.path.expanduser(path), "r") as stream:
                config = yaml.safe_load(stream)
        return config["github.com"] if config else {}



# Source: https://github.com/skypilot-org/skypilot/blob/feb52cf/sky/skylet/log_lib.py

import copy
import io
import multiprocessing.pool  # note: changed this
import os
import subprocess
import sys
from typing import List, Optional, Tuple, Union

from runhouse.resources.hardware.sky import subprocess_utils

class LineProcessor(object):
    """A processor for log lines."""

    def __enter__(self):
        pass

    def process_line(self, log_line):
        pass

    def __exit__(self, except_type, except_value, traceback):
        del except_type, except_value, traceback  # unused
        pass

class _ProcessingArgs:
    """Arguments for processing logs."""

    def __init__(self,
                 log_path: str,
                 stream_logs: bool,
                 start_streaming_at: str = '',
                 end_streaming_at: Optional[str] = None,
                 skip_lines: Optional[List[str]] = None,
                 replace_crlf: bool = False,
                 line_processor: Optional[LineProcessor] = None,
                 streaming_prefix: Optional[str] = None) -> None:
        self.log_path = log_path
        self.stream_logs = stream_logs
        self.start_streaming_at = start_streaming_at
        self.end_streaming_at = end_streaming_at
        self.skip_lines = skip_lines
        self.replace_crlf = replace_crlf
        self.line_processor = line_processor
        self.streaming_prefix = streaming_prefix


def process_subprocess_stream(proc, args: _ProcessingArgs) -> Tuple[str, str]:
    """Redirect the process's filtered stdout/stderr to both stream and file"""
    if proc.stderr is not None:
        # Asyncio does not work as the output processing can be executed in a
        # different thread.
        # selectors is possible to handle the multiplexing of stdout/stderr,
        # but it introduces buffering making the output not streaming.
        with multiprocessing.pool.ThreadPool(processes=1) as pool:
            err_args = copy.copy(args)
            err_args.line_processor = None
            stderr_fut = pool.apply_async(_handle_io_stream,
                                          args=(proc.stderr, sys.stderr,
                                                err_args))
            # Do not launch a thread for stdout as the rich.status does not
            # work in a thread, which is used in
            # log_utils.RayUpLineProcessor.
            stdout = _handle_io_stream(proc.stdout, sys.stdout, args)
            stderr = stderr_fut.get()
    else:
        stdout = _handle_io_stream(proc.stdout, sys.stdout, args)
        stderr = ''
    return stdout, stderr

def _handle_io_stream(io_stream, out_stream, args: _ProcessingArgs):
    """Process the stream of a process."""
    out_io = io.TextIOWrapper(io_stream,
                              encoding='utf-8',
                              newline='',
                              errors='replace',
                              write_through=True)

    start_streaming_flag = False
    end_streaming_flag = False
    streaming_prefix = args.streaming_prefix if args.streaming_prefix else ''
    line_processor = (LineProcessor()
                      if args.line_processor is None else args.line_processor)

    out = []
    with open(args.log_path, 'a') as fout:
        with line_processor:
            while True:
                line = out_io.readline()
                if not line:
                    break
                # start_streaming_at logic in processor.process_line(line)
                if args.replace_crlf and line.endswith('\r\n'):
                    # Replace CRLF with LF to avoid ray logging to the same
                    # line due to separating lines with '\n'.
                    line = line[:-2] + '\n'
                if (args.skip_lines is not None and
                        any(skip in line for skip in args.skip_lines)):
                    continue
                if args.start_streaming_at in line:
                    start_streaming_flag = True
                if (args.end_streaming_at is not None and
                        args.end_streaming_at in line):
                    # Keep executing the loop, only stop streaming.
                    # E.g., this is used for `sky bench` to hide the
                    # redundant messages of `sky launch` while
                    # saving them in log files.
                    end_streaming_flag = True
                if (args.stream_logs and start_streaming_flag and
                        not end_streaming_flag):
                    print(streaming_prefix + line,
                          end='',
                          file=out_stream,
                          flush=True)
                if args.log_path != '/dev/null':
                    fout.write(line)
                    fout.flush()
                line_processor.process_line(line)
                out.append(line)
    return ''.join(out)




def run_with_log(
    cmd: Union[List[str], str],
    log_path: str,
    *,
    require_outputs: bool = False,
    stream_logs: bool = False,
    start_streaming_at: str = '',
    end_streaming_at: Optional[str] = None,
    skip_lines: Optional[List[str]] = None,
    shell: bool = False,
    with_ray: bool = False,
    process_stream: bool = True,
    line_processor: Optional[LineProcessor] = None,
    streaming_prefix: Optional[str] = None,
    ray_job_id: Optional[str] = None,
    use_sudo: bool = False,
    **kwargs,
) -> Union[int, Tuple[int, str, str]]:
    """Runs a command and logs its output to a file.

    Args:
        cmd: The command to run.
        log_path: The path to the log file.
        stream_logs: Whether to stream the logs to stdout/stderr.
        require_outputs: Whether to return the stdout/stderr of the command.
        process_stream: Whether to post-process the stdout/stderr of the
            command, such as replacing or skipping lines on the fly. If
            enabled, lines are printed only when '\r' or '\n' is found.
        ray_job_id: The id for a ray job.
        use_sudo: Whether to use sudo to create log_path.

    Returns the returncode or returncode, stdout and stderr of the command.
      Note that the stdout and stderr is already decoded.
    """
    assert process_stream or not require_outputs, (
        process_stream, require_outputs,
        'require_outputs should be False when process_stream is False')

    log_path = os.path.expanduser(log_path)
    dirname = os.path.dirname(log_path)
    if use_sudo:
        # Sudo case is encountered when submitting
        # a job for Sky on-prem, when a non-admin user submits a job.
        subprocess.run(f'sudo mkdir -p {dirname}', shell=True, check=True)
        subprocess.run(f'sudo touch {log_path}; sudo chmod a+rwx {log_path}',
                       shell=True,
                       check=True)
        # Hack: Subprocess Popen does not accept sudo.
        # subprocess.Popen in local mode with shell=True does not work,
        # as it does not understand what -H means for sudo.
        shell = False
    else:
        os.makedirs(dirname, exist_ok=True)
    # Redirect stderr to stdout when using ray, to preserve the order of
    # stdout and stderr.
    stdout_arg = stderr_arg = None
    if process_stream:
        stdout_arg = subprocess.PIPE
        stderr_arg = subprocess.PIPE if not with_ray else subprocess.STDOUT
    with subprocess.Popen(cmd,
                          stdout=stdout_arg,
                          stderr=stderr_arg,
                          start_new_session=True,
                          shell=shell,
                          **kwargs) as proc:
        try:
            # The proc can be defunct if the python program is killed. Here we
            # open a new subprocess to gracefully kill the proc, SIGTERM
            # and then SIGKILL the process group.
            # Adapted from ray/dashboard/modules/job/job_manager.py#L154
            parent_pid = os.getpid()
            daemon_script = os.path.join(
                ### note: changed this
                os.path.dirname(os.path.abspath(__file__)),
                'subprocess_daemon.py')
            daemon_cmd = [
                'python3',
                daemon_script,
                '--parent-pid',
                str(parent_pid),
                '--proc-pid',
                str(proc.pid),
            ]
            # Bool use_sudo is true in the Sky On-prem case.
            # In this case, subprocess_daemon.py should run on the root user
            # and the Ray job id should be passed for daemon to poll for
            # job status (as `ray job stop` does not work in the
            # multitenant case).
            if use_sudo:
                daemon_cmd.insert(0, 'sudo')
                daemon_cmd.extend(['--local-ray-job-id', str(ray_job_id)])
            subprocess.Popen(
                daemon_cmd,
                start_new_session=True,
                # Suppress output
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                # Disable input
                stdin=subprocess.DEVNULL,
            )
            stdout = ''
            stderr = ''

            if process_stream:
                if skip_lines is None:
                    skip_lines = []
                # Skip these lines caused by `-i` option of bash. Failed to
                # find other way to turn off these two warning.
                # https://stackoverflow.com/questions/13300764/how-to-tell-bash-not-to-issue-warnings-cannot-set-terminal-process-group-and # pylint: disable=line-too-long
                # `ssh -T -i -tt` still cause the problem.
                skip_lines += [
                    'bash: cannot set terminal process group',
                    'bash: no job control in this shell',
                ]
                # We need this even if the log_path is '/dev/null' to ensure the
                # progress bar is shown.
                # NOTE: Lines are printed only when '\r' or '\n' is found.
                args = _ProcessingArgs(
                    log_path=log_path,
                    stream_logs=stream_logs,
                    start_streaming_at=start_streaming_at,
                    end_streaming_at=end_streaming_at,
                    skip_lines=skip_lines,
                    line_processor=line_processor,
                    # Replace CRLF when the output is logged to driver by ray.
                    replace_crlf=with_ray,
                    streaming_prefix=streaming_prefix,
                )
                stdout, stderr = process_subprocess_stream(proc, args)
            proc.wait()
            if require_outputs:
                return proc.returncode, stdout, stderr
            return proc.returncode
        except KeyboardInterrupt:
            # Kill the subprocess directly, otherwise, the underlying
            # process will only be killed after the python program exits,
            # causing the stream handling stuck at `readline`.
            subprocess_utils.kill_children_processes()
            raise



import time

import numpy as np

import pytest

import runhouse as rh
from runhouse.logger import get_logger

from tests.test_resources.test_modules.test_functions.test_function import (
    multiproc_np_sum,
)

TEMP_FILE = "my_file.txt"
TEMP_FOLDER = "~/runhouse-tests"

logger = get_logger(__name__)

UNIT = {"cluster": []}
LOCAL = {
    "cluster": [
        "docker_cluster_pwd_ssh_no_auth",
        "docker_cluster_pk_ssh_no_auth",
    ]
}
MINIMAL = {"cluster": ["local_launched_ondemand_aws_docker_cluster"]}
RELEASE = {
    "cluster": [
        "local_launched_ondemand_aws_docker_cluster",
        "ondemand_gcp_cluster",
        "ondemand_k8s_cluster",
        "ondemand_k8s_docker_cluster",
        "ondemand_aws_https_cluster_with_auth",
        "static_cpu_pwd_cluster",
    ]
}
MAXIMAL = {
    "cluster": [
        "docker_cluster_pwd_ssh_no_auth",
        "docker_cluster_pk_ssh_no_auth",
        "local_launched_ondemand_aws_docker_cluster",
        "ondemand_gcp_cluster",
        "ondemand_k8s_cluster",
        "ondemand_k8s_docker_cluster",
        "ondemand_aws_https_cluster_with_auth",
        "multinode_cpu_docker_conda_cluster",
        "static_cpu_pwd_cluster",
    ]
}


def do_printing_and_logging(steps=3):
    for i in range(steps):
        # Wait to make sure we're actually streaming
        time.sleep(1)
        print(f"Hello from the cluster stdout! {i}")
        logger.info(f"Hello from the cluster logs! {i}")
    return list(range(50))


def do_tqdm_printing_and_logging(steps=6):
    from tqdm.auto import tqdm  # progress bar

    progress_bar = tqdm(range(steps))
    for i in range(steps):
        # Wait to make sure we're actually streaming
        time.sleep(0.1)
        progress_bar.update(1)
    return list(range(50))


def test_stream_logs(cluster):
    print_fn = rh.function(fn=do_printing_and_logging, system=cluster)
    res = print_fn(stream_logs=True)
    # TODO [DG] assert that the logs are streamed
    assert res == list(range(50))


def test_get_from_cluster(cluster):
    print_fn = rh.function(fn=do_printing_and_logging, system=cluster)
    print(print_fn())
    res = print_fn.call.remote()
    assert isinstance(res, rh.Blob)
    assert res.name in cluster.keys()

    assert res.fetch() == list(range(50))
    res = cluster.get(res.name).resolved_state()
    assert res == list(range(50))


def test_put_and_get_on_cluster(cluster):
    test_list = list(range(5, 50, 2)) + ["a string"]
    cluster.put("my_list", test_list)
    ret = cluster.get("my_list")
    assert all(a == b for (a, b) in zip(ret, test_list))

    # Test that NOT_FOUND error is raised when key doesn't exist
    with pytest.raises(KeyError) as e:
        cluster.get("nonexistent_key", default=KeyError)
    assert "key nonexistent_key not found" in str(e.value)


class slow_numpy_array:
    def __init__(self, size=5):
        self.size = size
        self.arr = np.zeros(self.size)

    def slow_get_array(self):
        for i in range(self.size):
            time.sleep(1)
            logger.info(f"Hello from the cluster logs! {i}")
            self.arr[i] = i
            yield f"Hello from the cluster! {self.arr}"


@pytest.mark.parametrize("process", [None])
def test_stateful_generator(cluster, process):
    # We need this here just to make sure the "tests" module is synced over
    rh.function(fn=do_printing_and_logging, system=cluster)
    cluster.put("slow_numpy_array", slow_numpy_array(), process=process)
    for val in cluster.call("slow_numpy_array", "slow_get_array", stream_logs=True):
        assert val
        print(val)


def pinning_helper(key=None):
    from_obj_store = rh.here.get(key + "_inside")
    if from_obj_store:
        return "Found in obj store!"

    rh.blob(name=key + "_inside", data=["put within fn"] * 5).pin()
    return ["fn result"] * 3


def test_pinning_and_arg_replacement(cluster):
    cluster.clear()
    pin_fn = rh.function(pinning_helper).to(cluster)

    # First run should pin "run_pin" and "run_pin_inside"
    pin_fn.call.remote(key="run_pin", run_name="pinning_test")
    print(cluster.keys())
    assert cluster.get("pinning_test").fetch() == ["fn result"] * 3
    assert cluster.get("run_pin_inside").data == ["put within fn"] * 5

    # When we just ran with the arg "run_pin", we put a new pin called "pinning_test_inside"
    # from within the fn. Running again should return it.
    assert pin_fn("run_pin") == "Found in obj store!"

    put_pin_value = ["put_pin_value"] * 4
    cluster.put("put_pin_inside", put_pin_value)
    assert pin_fn("put_pin") == "Found in obj store!"


def test_put_resource(cluster):
    resource = rh.function(fn=do_printing_and_logging, system=cluster)
    cluster.put_resource(resource)
    assert cluster.call(resource.name, "config", stream_logs=True) == resource.config()
    assert cluster.get(resource.name).config() == resource.config()


def serialization_helper_1():
    import torch

    tensor = torch.zeros(100).cuda()
    rh.pin_to_memory("torch_tensor", tensor)


def serialization_helper_2():
    tensor = rh.get_pinned_object("torch_tensor")
    return tensor.device()  # Should succeed if array hasn't been serialized


@pytest.mark.skip
def test_pinning_to_gpu(k80_gpu_cluster):
    # Based on the following quirk having to do with Numpy objects becoming immutable if they're serialized:
    # https://docs.ray.io/en/latest/ray-core/objects/serialization.html#fixing-assignment-destination-is-read-only
    k80_gpu_cluster.delete()
    fn_1 = rh.function(serialization_helper_1).to(k80_gpu_cluster)
    fn_2 = rh.function(serialization_helper_2).to(k80_gpu_cluster)
    fn_1()
    fn_2()


def np_serialization_helper_1():
    print(time.time())
    rh.blob(data=np.zeros(100), name="np_arr").pin()


def np_serialization_helper_2():
    print(time.time())
    arr = rh.blob(name="np_arr").data
    arr[0] = 1
    return arr


def test_pinning_in_memory(cluster):
    # Based on the following quirk having to do with Numpy objects becoming immutable if they're serialized:
    # https://docs.ray.io/en/latest/ray-core/objects/serialization.html#fixing-assignment-destination-is-read-only
    cluster.clear()
    fn_1 = rh.function(np_serialization_helper_1).to(cluster)
    fn_1()
    fn_2 = rh.function(np_serialization_helper_2).to(cluster)
    res = fn_2()
    assert res[0] == 1
    assert res[1] == 0


def test_multiprocessing_streaming(cluster):
    re_fn = rh.function(multiproc_np_sum, system=cluster, env=["numpy"])
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands)
    assert res == [4, 6, 8, 10, 12]


def test_tqdm_streaming(cluster):
    # Note, this doesn't work properly in PyCharm due to incomplete
    # support for carriage returns in the PyCharm console.
    print_fn = rh.function(fn=do_tqdm_printing_and_logging).to(cluster, env=["tqdm"])
    res = print_fn(steps=40, stream_logs=True)
    assert res == list(range(50))



import os
import sys
from pathlib import Path


def subprocess_ray_fn_call_helper(pointers, args, kwargs, conn, ray_opts={}):
    def write_stdout(msg):
        conn.send((msg, "stdout"))

    def write_stderr(msg):
        conn.send((msg, "stderr"))

    sys.stdout.write = write_stdout
    sys.stderr.write = write_stderr
    abs_module_path = str(Path(pointers[0]).expanduser().resolve())

    ray_opts["runtime_env"] = ray_opts.get("runtime_env", {})
    ray_opts["runtime_env"]["env_vars"] = ray_opts["runtime_env"].get("env_vars", {})
    ray_opts["runtime_env"]["env_vars"]["RH_LOG_LEVEL"] = os.environ.get(
        "RH_LOG_LEVEL", "INFO"
    )
    if "PYTHONPATH" in ray_opts["runtime_env"]["env_vars"]:
        pp = ray_opts["runtime_env"]["env_vars"]["PYTHONPATH"]
        ray_opts["runtime_env"]["env_vars"]["PYTHONPATH"] = f"{abs_module_path}:{pp}"
    else:
        ray_opts["runtime_env"]["env_vars"]["PYTHONPATH"] = abs_module_path

    import ray

    ray.init(address="auto", **ray_opts)

    from runhouse.resources.module import Module

    (module_path, module_name, class_name) = pointers
    orig_fn = Module._get_obj_from_pointers(
        module_path, module_name, class_name, reload=False
    )
    try:
        res = orig_fn(*args, **kwargs)
        return res
    finally:
        ray.shutdown()

        # Send an EOFError over the pipe because for some reason .close is hanging
        conn.send((EOFError, None))
        conn.close()



import copy
import shlex
import subprocess
import time
from typing import Dict, Optional

from runhouse.constants import LOCALHOST, TUNNEL_TIMEOUT
from runhouse.globals import ssh_tunnel_cache

from runhouse.logger import get_logger
from runhouse.resources.hardware.sky.command_runner import SshMode
from runhouse.resources.hardware.sky.constants import DEFAULT_DOCKER_PORT
from runhouse.resources.hardware.utils import (
    _docker_ssh_proxy_command,
    _generate_ssh_control_hash,
    _ssh_base_command,
)

logger = get_logger(__name__)


class SshTunnel:
    def __init__(
        self,
        ip: str,
        ssh_user: str = None,
        ssh_private_key: str = None,
        ssh_control_name: Optional[str] = "__default__",
        ssh_proxy_command: Optional[str] = None,
        ssh_port: int = 22,
        disable_control_master: Optional[bool] = False,
        docker_user: Optional[str] = None,
        cloud: Optional[str] = None,
    ):
        """Initialize an ssh tunnel from a remote server to localhost

        Args:
            ip (str): The address of the server we are trying to port forward an address to our local machine with.
            ssh_user (str, optional): The SSH username to use for connecting to the remote server. Defaults to None.
            ssh_private_key (str, optional): The path to the SSH private key file. Defaults to None.
            ssh_control_name (str, optional): The name for the SSH control connection.
                Defaults to `"__default__"`.
            ssh_proxy_command (str, optional): The SSH proxy command to use for connecting to the remote
                server. Defaults to None.
            ssh_port (int, optional): The port on the remote machine where the SSH server is running. Defaults to 22.
            disable_control_master (bool, optional): Whether to disable SSH ControlMaster. Defaults to False.
            docker_user (str, optional): The Docker username to use if connecting through Docker. Defaults to None.
            cloud (str, optional): The cloud provider, if applicable. Defaults to None.
        """
        self.ip = ip
        self.ssh_port = ssh_port
        self.ssh_private_key = ssh_private_key
        self.ssh_control_name = (
            None
            if ssh_control_name is None
            else _generate_ssh_control_hash(ssh_control_name)
        )
        self.ssh_proxy_command = ssh_proxy_command
        self.disable_control_master = disable_control_master

        if docker_user:
            self.ssh_user = docker_user
            self.docker_ssh_proxy_command = _docker_ssh_proxy_command(
                ip, ssh_user, ssh_private_key
            )(["ssh"])

            if cloud != "kubernetes":
                self.ip = "localhost"
                self.ssh_port = DEFAULT_DOCKER_PORT
        else:
            self.ssh_user = ssh_user
            self.docker_ssh_proxy_command = None

        self.tunnel_proc = None

    def tunnel(self, local_port, remote_port):
        base_cmd = _ssh_base_command(
            address=self.ip,
            ssh_user=self.ssh_user,
            ssh_private_key=self.ssh_private_key,
            ssh_control_name=self.ssh_control_name,
            ssh_proxy_command=self.ssh_proxy_command,
            ssh_port=self.ssh_port,
            docker_ssh_proxy_command=self.docker_ssh_proxy_command,
            disable_control_master=self.disable_control_master,
            ssh_mode=SshMode.NON_INTERACTIVE,
            port_forward=[(local_port, remote_port)],
        )

        command = " ".join(base_cmd)
        logger.info(f"Running forwarding command: {command}")
        proc = subprocess.Popen(
            shlex.split(command),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # Wait until tunnel is formed by trying to create a socket in a loop

        start_time = time.time()
        while not is_port_in_use(local_port):
            time.sleep(0.1)
            if time.time() - start_time > TUNNEL_TIMEOUT:
                raise ConnectionError(
                    f"Failed to create tunnel from {local_port} to {remote_port} on {self.ip}"
                )

        # Set the tunnel process and ports to be cleaned up later
        self.tunnel_proc = proc
        self.local_bind_port = local_port
        self.remote_bind_port = remote_port

    def tunnel_is_up(self):
        # Try and do as much as we can to check that this is still alive and the port is still forwarded
        return self.local_bind_port is not None and is_port_in_use(self.local_bind_port)

    def __del__(self):
        self.terminate()

    def terminate(self):
        if self.tunnel_proc is not None:

            # Process keeping tunnel alive can only be killed with EOF
            self.tunnel_proc.stdin.close()

            # Remove port forwarding
            port_fwd_cmd = " ".join(
                _ssh_base_command(
                    address=self.ip,
                    ssh_user=self.ssh_user,
                    ssh_private_key=self.ssh_private_key,
                    ssh_control_name=self.ssh_control_name,
                    ssh_proxy_command=self.ssh_proxy_command,
                    ssh_port=self.ssh_port,
                    docker_ssh_proxy_command=self.docker_ssh_proxy_command,
                    disable_control_master=self.disable_control_master,
                    ssh_mode=SshMode.NON_INTERACTIVE,
                    port_forward=[(self.local_bind_port, self.remote_bind_port)],
                )
            )

            if "ControlMaster" in port_fwd_cmd:
                cancel_port_fwd = port_fwd_cmd.replace("-T", "-O cancel")
                logger.debug(f"Running cancel command: {cancel_port_fwd}")
                completed_cancel_cmd = subprocess.run(
                    shlex.split(cancel_port_fwd),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

                if completed_cancel_cmd.returncode != 0:
                    logger.warning(
                        f"Failed to cancel port forwarding from {self.local_bind_port} to {self.remote_bind_port}. "
                        f"Error: {completed_cancel_cmd.stderr}"
                    )

            self.tunnel_proc = None
            self.local_bind_port = None
            self.remote_bind_port = None


####################################################################################################
# Cache and retrieve existing SSH Runners that are set up for a given address and port
####################################################################################################
# TODO: Shouldn't the control master prevent new ssh connections from being created?


def is_port_in_use(port: int) -> bool:
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


def cache_existing_ssh_tunnel(address: str, ssh_port: int, tunnel: SshTunnel) -> None:
    ssh_tunnel_cache[(address, ssh_port)] = tunnel


def get_existing_sky_ssh_runner(address: str, ssh_port: int) -> Optional[SshTunnel]:
    if (address, ssh_port) in ssh_tunnel_cache:
        existing_tunnel = ssh_tunnel_cache.get((address, ssh_port))
        if existing_tunnel.tunnel_is_up():
            return existing_tunnel
        else:
            ssh_tunnel_cache.pop((address, ssh_port))
    else:
        return None


def ssh_tunnel(
    address: str,
    ssh_creds: Dict,
    local_port: int,
    ssh_port: int = 22,
    remote_port: Optional[int] = None,
    num_ports_to_try: int = 0,
    docker_user: Optional[str] = None,
    cloud: Optional[str] = None,
) -> SshTunnel:
    """Initialize an ssh tunnel from a remote server to localhost

    Args:
        address (str): The address of the server we are trying to port forward an address to our local machine with.
        ssh_creds (Dict): A dictionary of ssh credentials used to connect to the remote server.
        local_port (int): The port locally where we are attempting to bind the remote server address to.
        ssh_port (int): The port on the machine where the ssh server is running.
            This is generally port 22, but occasionally
            we may forward a container's ssh port to a different port
            on the actual machine itself (for example on a Docker VM). Defaults to 22.
        remote_port (int, optional): The port of the remote server
            we're attempting to port forward. Defaults to None.
        num_ports_to_try (int, optional): The number of local ports to attempt to bind to,
            starting at local_port and incrementing by 1 till we hit the max. Defaults to 0.
        docker_user (str, optional): The Docker username to use if connecting through Docker. Defaults to None.
        cloud (str, Optional): Cluster cloud, if an on-demand cluster.

    Returns:
        SshTunnel: The initialized tunnel.
    """

    # Debugging cmds (mac):
    # netstat -vanp tcp | grep 32300
    # lsof -i :32300
    # kill -9 <pid>

    # If remote_port isn't specified,
    # assume that the first attempted local port is
    # the same as the remote port on the server.
    remote_port = remote_port or local_port

    tunnel = get_existing_sky_ssh_runner(address, ssh_port)
    tunnel_address = address if not docker_user else "localhost"
    if (
        tunnel
        and tunnel.ip == tunnel_address
        and tunnel.remote_bind_port == remote_port
    ):
        logger.info(
            f"SSH tunnel on to server's port {remote_port} "
            f"via server's ssh port {ssh_port} already created with the cluster."
        )
        return tunnel

    while is_port_in_use(local_port):
        if num_ports_to_try < 0:
            raise Exception(
                f"Failed to create find open port after {num_ports_to_try} attempts"
            )

        logger.info(f"Port {local_port} is already in use. Trying next port.")
        local_port += 1
        num_ports_to_try -= 1

    ssh_credentials = copy.copy(ssh_creds)

    # Host could be a proxy specified in credentials or is the provided address
    host = ssh_credentials.pop("ssh_host", address)
    ssh_control_name = ssh_credentials.pop("ssh_control_name", f"{address}:{ssh_port}")

    tunnel = SshTunnel(
        ip=host,
        ssh_user=ssh_creds.get("ssh_user"),
        ssh_private_key=ssh_creds.get("ssh_private_key"),
        ssh_proxy_command=ssh_creds.get("ssh_proxy_command"),
        ssh_control_name=ssh_control_name,
        docker_user=docker_user,
        ssh_port=ssh_port,
        cloud=cloud,
    )
    tunnel.tunnel(local_port, remote_port)

    logger.debug(
        f"Successfully bound "
        f"{LOCALHOST}:{remote_port} via ssh port {ssh_port} "
        f"on remote server {address} "
        f"to {LOCALHOST}:{local_port} on local machine."
    )

    cache_existing_ssh_tunnel(address, ssh_port, tunnel)
    return tunnel



# Source: https://github.com/skypilot-org/skypilot/blob/feb52cf/sky/utils/subprocess_utils.py

from typing import Callable, List, Optional, Union

from runhouse.logger import get_logger

logger = get_logger(__name__)

class CommandError(Exception):
    """Raised when a command fails.

    Args:
        returncode: The returncode of the command.
        command: The command that was run.
        error_message: The error message to print.
        detailed_reason: The stderr of the command.
    """

    def __init__(self, returncode: int, command: str, error_msg: str,
                 detailed_reason: Optional[str]) -> None:
        self.returncode = returncode
        self.command = command
        self.error_msg = error_msg
        self.detailed_reason = detailed_reason
        message = (f'Command {command} failed with return code {returncode}.'
                   f'\n{error_msg}')
        super().__init__(message)


def handle_returncode(returncode: int,
                      command: str,
                      error_msg: Union[str, Callable[[], str]],
                      stderr: Optional[str] = None,
                      stream_logs: bool = True) -> None:
    """Handle the returncode of a command.

    Args:
        returncode: The returncode of the command.
        command: The command that was run.
        error_msg: The error message to print.
        stderr: The stderr of the command.
    """
    echo = logger.error if stream_logs else lambda _: None
    if returncode != 0:
        if stderr is not None:
            echo(stderr)

        if callable(error_msg):
            error_msg = error_msg()
        raise CommandError(returncode, command, error_msg, stderr)

def kill_children_processes(
        first_pid_to_kill: Optional[Union[int, List[Optional[int]]]] = None,
        force: bool = False):
    """Kill children processes recursively.

    We need to kill the children, so that
    1. The underlying subprocess will not print the logs to the terminal,
       after this program exits.
    2. The underlying subprocess will not continue with starting a cluster
       etc. while we are cleaning up the clusters.

    Args:
        first_pid_to_kill: Optional PID of a process, or PIDs of a series of
         processes to be killed first. If a list of PID is specified, it is
         killed by the order in the list.
         This is for guaranteeing the order of cleaning up and suppress
         flaky errors.
    """
    import psutil

    pid_to_proc = dict()
    child_processes = []
    if isinstance(first_pid_to_kill, int):
        first_pid_to_kill = [first_pid_to_kill]
    elif first_pid_to_kill is None:
        first_pid_to_kill = []

    def _kill_processes(processes: List[psutil.Process]) -> None:
        for process in processes:
            try:
                if force:
                    process.kill()
                else:
                    process.terminate()
            except psutil.NoSuchProcess:
                # The process may have already been terminated.
                pass

    parent_process = psutil.Process()
    for child in parent_process.children(recursive=True):
        if child.pid in first_pid_to_kill:
            pid_to_proc[child.pid] = child
        else:
            child_processes.append(child)

    _kill_processes([
        pid_to_proc[proc] for proc in first_pid_to_kill if proc in pid_to_proc
    ])
    _kill_processes(child_processes)



from runhouse.resources.secrets.provider_secrets.api_key_secret import ApiKeySecret


class AnthropicSecret(ApiKeySecret):
    """
    .. note::
            To create an AnthropicSecret, please use the factory method :func:`provider_secret`
            with ``provider="anthropic"``.
    """

    _PROVIDER = "anthropic"
    _DEFAULT_ENV_VARS = {"api_key": "ANTHROPIC_API_KEY"}

    @staticmethod
    def from_config(config: dict, dryrun: bool = False):
        return AnthropicSecret(**config, dryrun=dryrun)



from typing import Optional, Union

from runhouse.globals import rns_client
from runhouse.logger import get_logger
from runhouse.rns.utils.api import load_resp_content, ResourceAccess
from runhouse.servers.http.http_utils import username_from_token

logger = get_logger(__name__)


class AuthCache:
    # Maps a user's token to all the resources they have access to
    def __init__(self):
        self.CACHE = {}
        self.USERNAMES = {}

    def get_username(self, token: str) -> Optional[str]:
        """Get username associated with a particular token"""
        if token not in self.USERNAMES:
            username = username_from_token(token)
            if username is not None:
                self.USERNAMES[token] = username

        return self.USERNAMES.get(token)

    def lookup_access_level(
        self, token: str, resource_uri: str, refresh_cache=True
    ) -> Union[str, None]:
        """Get the access level of a particular resource for a user"""
        if token is None:
            return

        # Also add this user to the username cache
        self.get_username(token)

        if (token, resource_uri) in self.CACHE and not refresh_cache:
            return self.CACHE[(token, resource_uri)]

        if resource_uri.startswith("/"):
            resource_uri_to_send = resource_uri[1:].replace("/", ":")
        else:
            resource_uri_to_send = resource_uri.replace("/", ":")

        uri = f"{rns_client.api_server_url}/resource/{resource_uri_to_send}"
        resp = rns_client.session.get(
            uri,
            headers={"Authorization": f"Bearer {token}"},
        )

        if resp.status_code == 404:
            logger.error(
                f"Received [{resp.status_code}] from Den GET '{uri}': Resource not found: {load_resp_content(resp)}"
            )
            return

        if resp.status_code != 200:
            logger.error(
                f"Received [{resp.status_code}] from Den GET '{uri}': Failed to load access level for resource: {load_resp_content(resp)}"
            )
            return

        self.CACHE[(token, resource_uri)] = resp.json()["data"]["access_level"]

        return self.CACHE[(token, resource_uri)]

    def clear_cache(self, token: str = None):
        """Clear the server cache, If a token is specified, clear the cache for that particular user only"""
        if token is None:
            self.CACHE = {}
            self.USERNAMES = {}
        else:
            self.CACHE.pop(token, None)
            self.USERNAMES.pop(token, None)


async def averify_cluster_access(
    cluster_uri: str,
    token: str,
    access_level_required: ResourceAccess = None,
) -> bool:
    """Checks whether the user has access to the cluster.
    Note: A user with write access to the cluster or a cluster owner will have access to all other resources on
    the cluster by default."""
    from runhouse.globals import configs, obj_store

    # The logged-in user always has full access to the cluster. This is especially important if they flip on
    # Den Auth without saving the cluster. Note: The token saved in the cluster config is a hashed cluster token,
    # which may match the token provided in the request headers.
    if configs.token:
        if configs.token == token:
            return True

        if cluster_uri and rns_client.validate_cluster_token(
            cluster_token=token, cluster_uri=cluster_uri
        ):
            return True

    cluster_access_level = await obj_store.aresource_access_level(token, cluster_uri)

    if access_level_required is not None:
        return cluster_access_level == access_level_required

    return cluster_access_level in [ResourceAccess.WRITE, ResourceAccess.READ]



import fastapi

app = fastapi.FastAPI()


@app.get("/summer/{a}")
def summer(a: int, b: int):
    return a + b


@app.post("/my/deeply/{arg1}/nested/endpoint/{arg2}")
async def my_deeply_nested_async_endpoint(arg1: str, arg2: int, arg3: float):
    return arg1, arg2, arg3


@app.get("/my/streaming/endpoint")
def my_streaming_endpoint():
    for i in range(10):
        yield i


@app.get("/my/endpoint/with/optional/body/params/and/header")
def my_endpoint_with_optional_body_params_and_header(
    a: int = fastapi.Body(None),
    b: int = fastapi.Body(None),
    c: int = fastapi.Header(None),
):
    return a, b, c


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, port=8000)



import logging
import time

import pytest

from runhouse.resources.resource import Resource
from runhouse.servers.http.http_utils import deserialize_data, serialize_data
from runhouse.servers.obj_store import ObjStore

from tests.utils import init_remote_cluster_servlet_actor


@pytest.mark.servertest
class TestServlet:
    @pytest.mark.level("unit")
    def test_put_resource(self, test_servlet):
        resource = Resource(name="local-resource")
        state = {}
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aput_resource_local",
            data=serialize_data(
                (resource.config(condensed=False), state, resource.dryrun), "pickle"
            ),
            serialization="pickle",
        )
        assert resp.output_type == "result_serialized"
        assert deserialize_data(resp.data, resp.serialization) == resource.name

    @pytest.mark.level("unit")
    def test_put_obj_local(self, test_servlet):
        resource = Resource(name="local-resource")
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aput_local",
            key="key1",
            data=serialize_data(resource, "pickle"),
            serialization="pickle",
        )
        assert resp.output_type == "success"

    @pytest.mark.level("unit")
    def test_get_obj(self, test_servlet):
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aget_local",
            key="key1",
            default=KeyError,
            serialization="pickle",
            remote=False,
        )
        assert resp.output_type == "result_serialized"
        resource = deserialize_data(resp.data, resp.serialization)
        assert isinstance(resource, Resource)

    @pytest.mark.level("unit")
    def test_get_obj_remote(self, test_servlet):
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aget_local",
            key="key1",
            default=KeyError,
            serialization="pickle",
            remote=True,
        )
        assert resp.output_type == "config"
        resource_config = deserialize_data(resp.data, resp.serialization)
        assert isinstance(resource_config, dict)

    @pytest.mark.level("unit")
    def test_get_obj_does_not_exist(self, test_servlet):
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aget_local",
            key="abcdefg",
            default=KeyError,
            serialization="pickle",
            remote=False,
        )
        assert resp.output_type == "exception"
        error = deserialize_data(resp.data["error"], "pickle")
        assert isinstance(error, KeyError)

    @pytest.mark.level("local")
    def test_failed_cluster_servlet(self):
        import ray

        # need to initialize ray so the cluster servlet will be initialized when calling get_cluster_servlet.
        ray.init(
            ignore_reinit_error=True,
            logging_level=logging.ERROR,
            namespace="runhouse",
        )

        current_ip = ray.get_runtime_context().worker.node_ip_address

        cluster_config = {"name": "mocked_cluster"}

        init_remote_cluster_servlet_actor(
            current_ip=current_ip,
            servlet_name="invalid_cluster_servlet",
            cluster_config=cluster_config,
            runtime_env={
                "env_vars": {
                    "API_SERVER_URL": "https://api.run.house.invalid",
                    "RH_LOG_LEVEL": "DEBUG",
                }
            },
        )

        # wait for the async cluster status check thread to fail and kill the cluster servlet.
        time.sleep(5)

        with pytest.raises(ValueError) as error:

            ray.get_actor(name="invalid_cluster_servlet", namespace="runhouse")
        assert (
            str(error.value)
            == "Failed to look up actor with name 'invalid_cluster_servlet'. This could because 1. You are trying to look up a named actor you didn't create. 2. The named actor died. 3. You did not use a namespace matching the namespace of the actor."
        )



import codecs
import json
import re
import shutil
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import requests

from pydantic import BaseModel, field_validator

from runhouse.logger import get_logger

from runhouse.servers.obj_store import RunhouseStopIteration
from runhouse.utils import ClusterLogsFormatter

logger = get_logger(__name__)


class RequestContext(BaseModel):
    request_id: str
    token: Optional[str]


class InstallPackageParams(BaseModel):
    package_config: Dict[str, Any]
    conda_env_name: Optional[str] = None
    force_sync_local: bool = False


class RunBashParams(BaseModel):
    command: str
    node_ip_or_idx: Optional[Union[str, int]] = None
    process: Optional[str] = None
    require_outputs: bool = False
    run_name: Optional[str] = None


class ServerSettings(BaseModel):
    cluster_name: Optional[str] = None
    den_auth: Optional[bool] = None
    flush_auth_cache: Optional[bool] = None
    autostop_mins: Optional[int] = None
    status_check_interval: Optional[int] = None


class CreateProcessParams(BaseModel):
    name: str
    compute: Optional[Dict] = {}
    runtime_env: Optional[Dict] = {}
    env_vars: Optional[Dict] = {}


class SetEnvVarsParams(BaseModel):
    process_name: Optional[str] = None
    env_vars: Dict[str, str]


class CallParams(BaseModel):
    run_name: str
    data: Any = None
    serialization: Optional[str] = "none"
    stream_logs: Optional[bool] = False
    save: Optional[bool] = False
    remote: Optional[bool] = False


class PutResourceParams(BaseModel):
    serialized_data: Any
    serialization: Optional[str] = None
    process_name: Optional[str] = None


class PutObjectParams(BaseModel):
    key: str
    serialized_data: Any
    serialization: Optional[str] = None
    process_name: Optional[str] = None


class GetObjectParams(BaseModel):
    key: str
    serialization: Optional[str] = None
    remote: Optional[bool] = False


class KillProcessParams(BaseModel):
    process_name: str


class LogsParams(BaseModel):
    run_name: str
    node_ip_or_idx: Optional[Union[str, int]] = None
    process: Optional[str] = None
    key: Optional[str] = None
    serialization: Optional[str] = None


class RenameObjectParams(BaseModel):
    key: str
    new_key: str


class DeleteObjectParams(BaseModel):
    keys: List[str]


class Args(BaseModel):
    args: Optional[List[Any]]
    kwargs: Optional[Dict[str, Any]]


class Response(BaseModel):
    data: Any = None
    output_type: str
    serialization: Optional[str] = None


class OutputType:
    EXCEPTION = "exception"
    STDOUT = "stdout"
    STDERR = "stderr"
    SUCCESS = "success"  # No output
    CANCELLED = "cancelled"
    RESULT_SERIALIZED = "result_serialized"
    CONFIG = "config"


class FolderParams(BaseModel):
    path: str
    is_file: bool = False

    @field_validator("path", mode="before")
    def convert_path_to_string(cls, v):
        return str(v) if v is not None else v


class FolderLsParams(FolderParams):
    full_paths: Optional[bool] = True
    sort: Optional[bool] = False


class FolderGetParams(FolderParams):
    encoding: Optional[str] = None
    mode: Optional[str] = None


class FolderPutParams(FolderParams):
    contents: Any
    overwrite: Optional[bool] = False
    mode: Optional[str] = None
    serialization: Optional[str] = None


class FolderRmParams(FolderParams):
    contents: Optional[List] = None
    recursive: Optional[bool] = False


class FolderMvParams(FolderParams):
    dest_path: str
    overwrite: Optional[bool] = True

    @field_validator("dest_path", mode="before")
    def convert_path_to_string(cls, v):
        return str(v) if v is not None else v


def pickle_b64(picklable):
    import cloudpickle

    return codecs.encode(cloudpickle.dumps(picklable), "base64").decode()


def b64_unpickle(b64_pickled):
    import cloudpickle

    return cloudpickle.loads(codecs.decode(b64_pickled.encode(), "base64"))


def deserialize_data(data: Any, serialization: Optional[str]):
    if data is None:
        return None

    if serialization == "json":
        return json.loads(data)
    elif serialization == "pickle":
        return b64_unpickle(data)
    elif serialization is None or serialization == "none":
        return data
    else:
        raise ValueError(f"Invalid serialization type {serialization}")


def serialize_data(data: Any, serialization: Optional[str]):
    if data is None:
        return None

    if serialization == "json":
        try:
            return json.dumps(data)
        except TypeError:
            data["error"] = str(data["error"])
            return json.dumps(data)
    elif serialization == "pickle":
        return pickle_b64(data)
    elif serialization is None or serialization == "none":
        return data
    else:
        raise ValueError(f"Invalid serialization type {serialization}")


def handle_exception_response(
    exception: Exception, traceback: str, serialization="pickle", from_http_server=False
):
    from fastapi import HTTPException
    from ray.exceptions import RayTaskError

    if not (
        isinstance(exception, RunhouseStopIteration)
        or isinstance(exception, StopIteration)
        or isinstance(exception, GeneratorExit)
        or isinstance(exception, StopAsyncIteration)
    ):
        logger.exception(exception)

    # We need to be selective about when we convert errors to HTTPExceptions because HTTPExceptions are not
    # picklable and can't be passed over the actor boundary. We should only be converting in the HTTPServer right
    # before we send the response back to the client from the http_server.
    if isinstance(exception, RayTaskError) and from_http_server:
        cause = exception.args[0]
        detail = cause.args[0] if cause.args else ""
        if isinstance(cause, PermissionError):
            if "No Runhouse token provided." in detail:
                raise HTTPException(status_code=401, detail=detail)
            if "Unauthorized access to resource" in detail:
                raise HTTPException(status_code=403, detail=detail)
        if isinstance(cause, ValueError):
            if "Invalid serialization type" in detail:
                raise HTTPException(status_code=400, detail=detail)

    if isinstance(exception, PermissionError):
        # If we're raising from inside the obj store, we should still raise this error so the HTTPServer can
        # catch it and convert it to an HTTPException.
        raise exception

    if serialization not in ["json", "pickle"] or isinstance(exception, HTTPException):
        # If the exception is an HTTPException, we should let it flow back through the server so it gets
        # handled by FastAPI's exception handling and returned to the user with the correct status code
        raise exception

    if isinstance(exception, RunhouseStopIteration):
        exception = StopIteration()

    exception_data = {
        "error": serialize_data(exception, serialization),
        "traceback": traceback,
        "exception_as_str": str(exception),
    }
    return Response(
        output_type=OutputType.EXCEPTION,
        data=exception_data,
        serialization=serialization,
    )


def username_from_token(token: str) -> Union[str, None]:
    """Get the username from the provided cluster subtoken."""
    from runhouse.globals import rns_client
    from runhouse.rns.utils.api import read_resp_data

    resp = requests.get(
        f"{rns_client.api_server_url}/user",
        headers={"Authorization": f"Bearer {token}"},
    )
    if resp.status_code != 200:
        return None

    user_data = read_resp_data(resp)
    username = user_data.get("username")
    if not username:
        return None

    return username


def get_token_from_request(request):
    auth_headers = auth_headers_from_request(request)
    return auth_headers.split("Bearer ")[-1] if auth_headers else None


def auth_headers_from_request(request):
    return request.headers.get("Authorization", "")


def load_current_cluster_rns_address():
    from runhouse.resources.hardware import _current_cluster, _get_cluster_from

    current_cluster = _get_cluster_from(_current_cluster("config"))
    return current_cluster.rns_address if current_cluster else None


def handle_response(
    response_data: Dict[Any, Any],
    output_type: OutputType,
    err_str: str,
    log_formatter: ClusterLogsFormatter,
):
    system_color, reset_color = log_formatter.format_server_log(output_type)

    if output_type == OutputType.RESULT_SERIALIZED:
        return deserialize_data(response_data["data"], response_data["serialization"])
    elif output_type == OutputType.CONFIG:
        # No need to unpickle since this was just sent as json
        return response_data["data"]
    elif output_type == OutputType.CANCELLED:
        raise RuntimeError(f"{err_str}: task was cancelled")
    elif output_type == OutputType.SUCCESS:
        return
    elif output_type == OutputType.EXCEPTION:
        # Here for compatibility before we stopped serializing the traceback
        if not isinstance(response_data["data"], dict):
            exception_dict = deserialize_data(
                response_data["data"], response_data["serialization"]
            )
            fn_exception = exception_dict["error"]
            fn_traceback = exception_dict["traceback"]
        else:
            fn_traceback = response_data["data"]["traceback"]
            fn_exception = None
            fn_exception_as_str = response_data["data"].get("exception_as_str", None)
            try:
                fn_exception = deserialize_data(
                    response_data["data"]["error"], response_data["serialization"]
                )
            except Exception as e:
                logger.error(
                    f"{system_color}{err_str}: Failed to unpickle exception. Please check the logs for more "
                    f"information.{reset_color}"
                )
                if fn_exception_as_str:
                    logger.error(
                        f"{system_color}{err_str} Exception as string: {fn_exception_as_str}{reset_color}"
                    )
                logger.error(f"{system_color}Traceback: {fn_traceback}{reset_color}")
                raise e
        if not (
            isinstance(fn_exception, StopIteration)
            or isinstance(fn_exception, GeneratorExit)
            or isinstance(fn_exception, StopAsyncIteration)
        ):
            logger.error(f"{system_color}{err_str}: {fn_exception}{reset_color}")
            logger.error(f"{system_color}Traceback: {fn_traceback}{reset_color}")
        raise fn_exception
    elif output_type == OutputType.STDOUT:
        res = response_data["data"]
        # Regex to match tqdm progress bars
        tqdm_regex = re.compile(r"(.+)%\|(.+)\|\s+(.+)/(.+)")
        for line in res:
            if tqdm_regex.match(line):
                # tqdm lines are always preceded by a \n, so we can use \x1b[1A to move the cursor up one line
                # For some reason, doesn't work in PyCharm's console, but works in the terminal
                print(
                    f"{system_color}\x1b[1A\r" + line + reset_color, end="", flush=True
                )
            else:
                print(system_color + line + reset_color, end="", flush=True)
    elif output_type == OutputType.STDERR:
        res = response_data["data"]
        print(system_color + res + reset_color, file=sys.stderr)


###########################
#### Folder Operations ####
###########################


def resolve_folder_path(path: str):
    return (
        None
        if path is None
        else Path(path).expanduser()
        if path.startswith("~")
        else Path(path).resolve()
    )


def folder_mkdir(path: Path):
    if not path.parent.is_dir():
        raise ValueError(
            f"Parent path {path.parent} does not exist or is not a directory"
        )

    path.mkdir(parents=True, exist_ok=True)

    return Response(output_type=OutputType.SUCCESS)


def folder_get(path: Path, encoding: str = None, mode: str = None):
    from fastapi import HTTPException

    mode = mode or "rb"
    binary_mode = "b" in mode

    if not path.exists():
        raise HTTPException(status_code=404, detail=f"Path {path} does not exist")

    try:
        with open(path, mode=mode, encoding=encoding) as f:
            file_contents = f.read()

    except FileNotFoundError:
        raise HTTPException(status_code=404, detail=f"File {path} not found")

    except PermissionError:
        raise HTTPException(
            status_code=403, detail=f"Permission denied for file in path {path}"
        )

    except Exception as e:
        raise HTTPException(
            status_code=500, detail=f"Error reading file {path}: {str(e)}"
        )

    if binary_mode and isinstance(file_contents, bytes):
        file_contents = file_contents.decode()

    return Response(
        data=file_contents,
        output_type=OutputType.RESULT_SERIALIZED,
        serialization=None,
    )


def folder_put(
    path: Path,
    contents: Dict[str, Any],
    overwrite: bool,
    mode: str = None,
    serialization: str = None,
):
    from fastapi import HTTPException

    mode = mode or "wb"

    if contents and not isinstance(contents, dict):
        raise HTTPException(
            status_code=422,
            detail="`contents` argument must be a dict mapping filenames to file-like objects",
        )

    if overwrite is False:
        existing_files = {str(item.name) for item in path.iterdir()}
        intersection = existing_files.intersection(set(contents.keys()))
        if intersection:
            raise HTTPException(
                status_code=409,
                detail=f"File(s) {intersection} already exist(s) at path: {path}",
            )

    path.mkdir(parents=True, exist_ok=True)

    for filename, file_obj in contents.items():
        binary_mode = "b" in mode

        if serialization:
            file_obj = serialize_data(file_obj, serialization)

        if binary_mode and not isinstance(file_obj, bytes):
            file_obj = file_obj.encode()

        file_path = path / filename
        if not overwrite and file_path.exists():
            raise HTTPException(
                status_code=409, detail=f"File {file_path} already exists"
            )

        try:
            with open(file_path, mode) as f:
                f.write(file_obj)
        except Exception as e:
            HTTPException(
                status_code=500,
                detail=f"Failed to write file with mode '{mode}': {str(e)}",
            )

    return Response(output_type=OutputType.SUCCESS)


def folder_ls(path: Path, full_paths: bool, sort: bool):
    from fastapi import HTTPException

    if not path.exists():
        raise HTTPException(status_code=404, detail=f"Path {path} does not exist")

    if not path.is_dir():
        raise HTTPException(status_code=400, detail=f"Path {path} is not a directory")

    paths = [p for p in path.iterdir()]

    # Sort the paths by modification time if sort is True
    if sort:
        paths.sort(key=lambda p: p.stat().st_mtime, reverse=True)

    # Convert paths to strings and format them based on full_paths
    if full_paths:
        files = [str(p.resolve()) for p in paths]
    else:
        files = [p.name for p in paths]

    return Response(
        data=files,
        output_type=OutputType.RESULT_SERIALIZED,
        serialization=None,
    )


def folder_rm(path: Path, contents: List[str], recursive: bool):
    from fastapi import HTTPException

    if contents:
        from runhouse import Folder

        try:
            Folder._delete_contents(contents, path, recursive)
        except ValueError as e:
            raise HTTPException(
                status_code=400,
                detail=str(e),
            )

        return Response(output_type=OutputType.SUCCESS)

    if not path.is_dir():
        path.unlink()
        return Response(output_type=OutputType.SUCCESS)

    if recursive:
        shutil.rmtree(path)
        return Response(output_type=OutputType.SUCCESS)

    items = list(path.iterdir())
    if not items:
        # Remove the empty directory
        path.rmdir()
        return Response(output_type=OutputType.SUCCESS)

    # Remove file contents, but not the directory itself (since recursive not set to `True`)
    for item in items:
        if item.is_file():
            item.unlink()
        else:
            raise HTTPException(
                status_code=400,
                detail=f"Folder {item} found in {path}, recursive is set to `False`",
            )

    return Response(output_type=OutputType.SUCCESS)


def folder_mv(src_path: Path, dest_path: str, overwrite: bool):
    from fastapi import HTTPException

    dest_path = resolve_folder_path(dest_path)

    if not src_path.exists():
        raise HTTPException(
            status_code=404, detail=f"The source path {src_path} does not exist"
        )

    if not overwrite and dest_path.exists():
        raise HTTPException(
            status_code=409,
            detail=f"The destination path {dest_path} already exists. Set `overwrite` to `True` to "
            f"overwrite the destination path",
        )

    # Create the destination directory if it doesn't exist and overwrite is set to `True`
    dest_path.parent.mkdir(parents=True, exist_ok=overwrite)

    # Move the directory
    shutil.move(str(src_path), str(dest_path))

    return Response(output_type=OutputType.SUCCESS)


def folder_exists(path: Path):
    folder_exists_resp = path.exists()
    if not path.is_file():
        folder_exists_resp = folder_exists_resp and path.is_dir()
    return Response(
        data=folder_exists_resp,
        output_type=OutputType.RESULT_SERIALIZED,
        serialization=None,
    )
